#!/usr/bin/env python

""" Broad collection of utility functions and constants.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import functools
import logging
import operator
from typing import Any, Mapping, Sequence, Union

import numpy as np

# Setup logger
logger = logging.getLogger(__name__)

# Small value - epsilon
# For use to offset from bin edges when finding bins for use with SetRange()
# NOTE: sys.float_info.epsilon is too small in some cases and thus should be avoided
epsilon = 1e-5

###################
# Utility functions
###################
def moving_average(arr: np.ndarray, n: int = 3) -> np.ndarray:
    """ Calculate the moving overage over an array.

    Algorithm from: https://stackoverflow.com/a/14314054

    Args:
        arr (np.ndarray): Array over which to calculate the moving average.
        n (int): Number of elements over which to calculate the moving average. Default: 3
    Returns:
        np.ndarray: Moving average calculated over n.
    """
    ret = np.cumsum(arr, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def recursive_getattr(obj: Any, attr: str, *args: Any) -> Any:
    """ Recursive ``getattar``.

    This can be used as a drop in for the standard ``getattr(...)``. Credit to:
    https://stackoverflow.com/a/31174427

    Args:
        obj: Object to retrieve the attribute from.
        attr: Name of the attribute, with each successive attribute separated by a ".".
    Returns:
        The requested attribute. (Same as ``getattr``).
    Raises:
        AttributeError: If the attribute was not found and no default was provided. (Same as ``getattr``).
    """
    def _getattr(obj_recurse: Any, attr_recurse: str) -> Any:
        return getattr(obj_recurse, attr_recurse, *args)
    return functools.reduce(_getattr, [obj] + attr.split('.'))

def recursive_setattr(obj: Any, attr: str, val: Any) -> Any:
    """ Recusrive ``setattr``.

    This can be used as a drop in for the standard ``setattr(...)``. Credit to:
    https://stackoverflow.com/a/31174427

    Args:
        obj: Object to retrieve the attribute from.
        attr: Name of the attribute, with each successive attribute separated by a ".".
        value: Value to set the attribute to.
    Returns:
        The requested attribute. (Same as ``getattr``).
    Raises:
        AttributeError: If the attribute was not found and no default was provided. (Same as ``getattr``).
    """
    pre, _, post = attr.rpartition('.')
    return setattr(recursive_getattr(obj, pre) if pre else obj, post, val)

def recursive_getitem(d: Mapping[str, Any], keys: Union[str, Sequence[str]]) -> Any:
    """ Recursively retrieve an item from a nested dict.

    Credit to: https://stackoverflow.com/a/52260663

    Args:
        d: Mapping of strings to objects.
        keys: Names of the keys under which the object is stored. Can also just be a single string.
    Returns:
        The object stored under the keys.
    Raises:
        KeyError: If one of the keys isnt' found.
    """
    # If only a string, then just just return the item
    if isinstance(keys, str):
        return d[keys]
    else:
        return functools.reduce(operator.getitem, keys, d)



#!/usr/bin/env python3

""" Integration of functionality in the fit modules.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import abc
import logging
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Type, TypeVar, Union

import iminuit
import numpy as np

from pachyderm import generic_class
from pachyderm.fit import base, cost_function

if TYPE_CHECKING:
    from pachyderm import histogram, yaml

logger = logging.getLogger(__name__)

# Typing
T_FitArguments = Dict[str, Union[bool, float, Tuple[float, float]]]

_T_Fit = TypeVar("_T_Fit", bound = "Fit")

class Fit(abc.ABC, generic_class.EqualityMixin):
    """ Class to direct fitting a histogram to a fit function.

    This allows us to easily store the fit function right alongside the minimization.

    Args
        use_log_likelihood: True if the log likelihood cost function should be used.
        fit_options: Options to use for the fit, such as the range in which the data will be fit.
            It's up to the derived class to determine how to use this information.
        user_arguments: User arguments for the fit. Default: None.

    Attributes:
        use_log_likelihood: True if the log likelihood cost function should be used.
        fit_options: Options to use for the fit, such as the range in which the data will be fit.
            It's up to the derived class to determine how to use this information.
        user_arguments: User arguments for the fit. Default: None.
        fit_function: Function to be fit.
        fit_result: Result of the fit. Only valid after the fit has been performed.
    """
    def __init__(self, use_log_likelihood: bool, fit_options: Optional[Dict[str, Any]] = None,
                 user_arguments: Optional[T_FitArguments] = None):
        # Validation
        if user_arguments is None:
            user_arguments = {}
        if fit_options is None:
            fit_options = {}

        # Store main parameters
        self.use_log_likelihood = use_log_likelihood
        self.fit_options = fit_options
        self.user_arguments: T_FitArguments = user_arguments
        self.fit_function: Callable[..., float]
        self.fit_result: base.FitResult

        # Create the cost function based on the fit parameters.
        self._cost_func: Type[cost_function.DataComparisonCostFunction]
        if use_log_likelihood:
            self._cost_func = cost_function.BinnedLogLikelihood
        else:
            self._cost_func = cost_function.BinnedChiSquared

        # Check fit initialization.
        self._post_init_validation()

    @abc.abstractmethod
    def _post_init_validation(self) -> None:
        """ Validate that the fit object was setup properly.

        This can be any method that the user devises to ensure that
        all of the information needed for the fit is available.

        Args:
            None.
        Returns:
            None.
        """
        ...

    @abc.abstractmethod
    def _setup(self, h: "histogram.Histogram1D") -> Tuple["histogram.Histogram1D", T_FitArguments]:
        """ Setup the histogram and arguments for the fit.

        Args:
            h: Background subtracted histogram to be fit.
        Returns:
            Histogram to use for the fit, default arguments for the fit. Note that the histogram may be range
                restricted or otherwise modified here.
        """
        ...

    def _create_cost_function(self, h: "histogram.Histogram1D") -> cost_function.DataComparisonCostFunction:
        """ Create the cost function from the data and stored parameters.

        Args:
            h: Data to be used for the fit.
        Returns:
            The created cost function.
        """
        return self._cost_func(f = self.fit_function, data = h)

    def __call__(self, *args: float, **kwargs: float) -> float:
        """ Call the fit function.

        This is provided for convenience. This way, we can easily evaluate the function while
        still storing the information necessary to perform the entire fit.

        Args:
            args: Arguments to pass to the fit function.
            kwargs: Arguments to pass to the fit function.
        Returns:
            The fit function called with these arguments.
        """
        return self.fit_function(*args, **kwargs)

    def calculate_errors(self, x: Optional[np.ndarray] = None) -> np.ndarray:
        """ Calculate the errors on the fit function for the given x values.

        Args:
            x: x values where the fit function error should be evaluated. If not specified,
                the x values over which the fit was performed will be used.
        Returns:
            The fit function error calculated at each x value.
        """
        if x is None:
            x = self.fit_result.x
        return base.calculate_function_errors(
            func = self.fit_function,
            fit_result = self.fit_result,
            x = x,
        )

    def fit(self, h: "histogram.Histogram1D", user_arguments: Optional[T_FitArguments] = None) -> base.FitResult:
        """ Fit the given histogram to the stored fit function using iminuit.

        The fit errors will be automatically calculated if possible. It is possible if the fit function
        accessible through the cost function is not an added PDF.

        Args:
            h: Histogram to be fit.
            user_arguments: Additional user arguments (beyond those already specified when the object
                was created). They will override the already specified options. Default: None.
        Returns:
            Result of the fit. The user is responsible for storing it in the fit.
        """
        # Validation
        if user_arguments is None:
            user_arguments = {}

        # Setup the fit and the cost function
        hist_for_fit, user_fit_arguments = self._setup(h = h)
        # Update the default user arguments provided in setup by...
        # ... The stored user arguments when the object was created.
        user_fit_arguments.update(self.user_arguments)
        # ... The user arguments passed to the fit function.
        user_fit_arguments.update(user_arguments)
        # Then create the cost function according to the parameters and (potentially restricted) data.
        cost_func = self._create_cost_function(h = hist_for_fit)

        # Perform the fit by minimizing the chi squared
        fit_result, _ = fit_with_minuit(
            cost_func = cost_func, minuit_args = user_fit_arguments, x = hist_for_fit.x,
            use_minos = self.fit_options.get("minos", False),
        )

        # Calculate the fit result only if requested.
        if self.fit_options.get("calculate_errors", False):
            fit_result.errors = self.calculate_errors(hist_for_fit.x)

        return fit_result

    @classmethod
    def to_yaml(cls: Type[_T_Fit], representer: "yaml.Representer",
                obj: _T_Fit) -> "yaml.ruamel.yaml.nodes.SequenceNode":
        """ Encode YAML representation.

        Since YAML won't handle function very nicely, we convert them to strings and then check them
        on conversion from YAML as a cross check that recreating the object hasn't gone wrong. This
        also requires all functions to be recreatable on object initialization.

        Args:
            representer: Representation from YAML.
            data: Fit function to be converted to YAML.
        Returns:
            YAML representation of the fit object.
        """
        # ``RoundTripRepresenter`` doesn't represent objects directly, so we grab a dict of the members to
        # store them in YAML.
        # NOTE: We must make a copy of the vars. Otherwise, the original fit object will be modified.
        members = dict(vars(obj))
        # We can't store unbound functions, so we instead set it to the function name (we won't really
        # use this name to try to directly recreate the function when the object is being loaded from YAML,
        # but it's useful to store what was used, and we can at least warn if it changed).
        members["fit_function"] = obj.fit_function.__name__
        members["_cost_func"] = obj._cost_func.__name__
        representation = representer.represent_mapping(
            f"!{cls.__name__}", members
        )

        # Finally, return the represented object.
        return representation

    @classmethod
    def from_yaml(cls: Type[_T_Fit], constructor: "yaml.Constructor",
                  data: "yaml.ruamel.yaml.nodes.MappingNode") -> _T_Fit:
        """ Decode YAML representation.

        Args:
            constructor: Constructor from the YAML object.
            node: YAML Mapping node representing the fit object.
        Returns:
            The fit object constructed from the YAML specified values.
        """
        # First, we construct the class member objects.
        members = {
            constructor.construct_object(key_node): constructor.construct_object(value_node)
            for key_node, value_node in data.value
        }
        # Then we deal with members which require special handling:
        # The fit result isn't set through the constructor, so we grab it and then
        # set it after creating the object.
        fit_result = members.pop("fit_result", None)
        # The fit function will be set in the fit constructor, so we don't need to use this
        # value to setup the object. However, since this contains the name of the function,
        # we can use it to check if the name of the function that is set in the constructor
        # is the same as the one that we stored. (If they are different, this isn't necessarily
        # a problem, as we sometimes rename functions, but regardless it's good to be notified
        # if that's the case).
        function_names = {}
        for attr_name in ["fit_function", "_cost_func"]:
            function_names[attr_name] = members.pop(attr_name)

        # Finally, create the object and set the properties as needed.
        obj = cls(**members)
        # The fit result may not have been defined yet, so only set it if it has.
        if fit_result is not None:
            obj.fit_result = fit_result
        # Sanity checks on fit and cost function names (see above).
        for attr_name, function_name in function_names.items():
            if function_name != getattr(obj, attr_name).__name__:
                logger.warning(
                    "The stored '{attr_name}' function name from YAML doesn't match the name of the function"
                    " created in the fit object."
                    f" Stored name: {function_name}, object created fit function: {getattr(obj, attr_name).__name__}."
                    " This may indicate a problem (but is fine if the same function was just renamed)."
                )

        # Now that the object is fully constructed, we can return it.
        return obj

def _validate_minuit_args(cost_func: Union[cost_function.CostFunctionBase, cost_function.SimultaneousFit],
                          minuit_args: T_FitArguments) -> None:
    """ Validate the arguments provided for Minuit.

    Checks that there are sufficient and valid arguments for each parameter in the fit function.

    Args:
        cost_func: Cost function to be used with Minuit.
        minuit_args: Arguments for minuit. Need to set the initial value, limits, and error (step)
            of each parameter.
    Returns:
        None. An exception is raised if there's a problem with any of the arguments.
    """
    # Need the parameters in the function to check the arguments. Recall that "x" is skipped because it's a cost
    # function (which is useful, because we want to skip x since it's not a parameter).
    parameters = iminuit.describe(cost_func)
    # Loop over the available parameters because each one needs to be specified in the Minuit args.
    for p in parameters:
        if p in minuit_args:
            if f"limit_{p}" not in minuit_args:
                raise ValueError(f"Limits on parameter '{p}' must be specified.")
            if f"error_{p}" not in minuit_args:
                raise ValueError(f"Initial error on parameter '{p}' must be specified.")
            # If p and limit_p and error_p were specified, then the parameter is fully specified.
        elif p.replace("fix", "") in minuit_args:
            # The parameter is fixed and therefore needs no further specification.
            pass
        else:
            # The parameter wasn't specified.
            raise ValueError(f"Parameter '{p}' must be specified in the fit arguments.")

def fit_with_minuit(cost_func: Union[cost_function.CostFunctionBase, cost_function.SimultaneousFit],
                    minuit_args: T_FitArguments, x: np.ndarray,
                    use_minos: Optional[bool] = False) -> Tuple[base.FitResult, iminuit.Minuit]:
    """ Perform a fit using the given cost function with Minuit.

    Args:
        cost_func: Cost function to be used with Minuit.
        minuit_args: Arguments for minuit. Need to set the initial value, limits, and error (step)
            of each parameter.
        x: x value(s) where the fit is evaluated, which will be stored in the fit result.
        use_minos: Calculate MINOS errors. They have to be accessed through the Minuit object. Default: False.
    Returns:
        (fit_result, Minuit object): The fit result extracts values from the Minuit object, but
            the Minuit object is also returned for good measure.
    """
    # Validation
    # Will raise an exception if there are invalid arguments.
    _validate_minuit_args(cost_func = cost_func, minuit_args = minuit_args)
    # Set the error definition.
    # We check if it's set to the allow the user to override if they are so inclined.
    # (Overriding it should be pretty rare).
    if "errordef" not in minuit_args:
        # Log likelihood cost functions needs an errordef of 0.5 to scale the errors properly, while 1 should
        # be used for chi squared cost functions.
        error_def = 1.0
        if isinstance(cost_func, (cost_function.LogLikelihood, cost_function.BinnedLogLikelihood)):
            error_def = 0.5
        # Store the value.
        minuit_args["errordef"] = error_def

    # Perform the fit
    minuit = iminuit.Minuit(cost_func, **minuit_args)
    # Improve minimization reliability.
    minuit.set_strategy(2)
    minuit.migrad()
    # Just in case (doesn't hurt anything, but may help in a few cases).
    minuit.hesse()
    if use_minos:
        minuit.minos()

    # Check that the fit is actually good
    if not minuit.migrad_ok():
        raise base.FitFailed("Minimization failed! The fit is invalid!")
    # Check covariance matrix accuracy. We need to check it explicitly because It appears that it is not
    # included in the migrad_ok status check.
    if not minuit.matrix_accurate():
        raise base.FitFailed("Covariance matrix is inaccurate! The fit is invalid!")

    # Create the fit result and calculate the errors.
    fit_result = base.FitResult.from_minuit(minuit, cost_func, x)
    # We can calculate the fit errors if the cost function has a single function.
    # If it's a simultaneous fit, it's unclear how best this should be handled. Perhaps it could
    # be unraveled and summed, but it's not obvious that that's the best approach. More likely,
    # one only wants the errors for an individual cost function, so we leave that to the user.
    # We use getattr instead of hasattr to help out mypy
    if isinstance(cost_func, cost_function.CostFunctionBase):
        errors = base.calculate_function_errors(cost_func.f, fit_result, x)
    else:
        errors = []
    fit_result.errors = errors

    return fit_result, minuit



#!/usr/bin/env python3

""" Tests for the cost functions module.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import logging
from typing import Any, Dict, Tuple, Union

import numpy as np
import pytest
import scipy.integrate

import pachyderm.fit.base as fit_base
import pachyderm.fit.integration as fit_integration
from pachyderm import histogram
from pachyderm.fit import cost_function
from pachyderm.typing_helpers import Hist

logger = logging.getLogger(__name__)

def func_1(x: float, a: float, b: float) -> float:
    """ Test function. """
    return x + a + b

def func_2(x: float, c: float, d: float) -> float:
    """ Test function 2. """
    return x + c + d

def test_integration(logging_mixin: Any) -> None:
    """ Test our implementation of the Simpson 3/8 rule, along with some other integration methods. """
    # Setup
    f = func_1
    h = histogram.Histogram1D(
        bin_edges = np.array([0, 1, 2]), y = np.array([0, 1]), errors_squared = np.array([1, 2])
    )
    args = [0, 0]

    integral = cost_function._simpson_38(f, h.bin_edges, *args)
    # Evaluate at the bin center
    expected = np.array([f(i, *args) for i in h.x])
    np.testing.assert_allclose(integral, expected)

    # Compare against our quad implementation
    integral_quad = cost_function._quad(f, h.bin_edges, *args)
    np.testing.assert_allclose(integral, integral_quad)

    # Also compare against probfit and scipy for good measure
    probfit = pytest.importorskip("probfit")
    expected_probfit = []
    expected_scipy = []
    expected_quad = []
    for i in h.bin_edges[1:]:
        # Assumes uniform bin width
        expected_probfit.append(probfit.integrate1d(f, (i - 1, i), 1, tuple(args)))
        scipy_x = np.linspace(i - 1, i, 5)
        expected_scipy.append(scipy.integrate.simps(y = f(scipy_x, *args), x = scipy_x))
        res, _ = scipy.integrate.quad(f, i - 1, i, args = tuple(args))
        expected_quad.append(res)

    np.testing.assert_allclose(integral, expected_probfit)
    np.testing.assert_allclose(integral, expected_scipy)
    np.testing.assert_allclose(integral, expected_quad)

def test_chi_squared(logging_mixin: Any) -> None:
    """ Test the chi squared calculation. """
    # Setup
    h = histogram.Histogram1D(
        bin_edges = np.array(np.arange(-0.5, 5.5)), y = np.array(np.ones(5)), errors_squared = np.ones(5),
    )
    chi_squared = cost_function.ChiSquared(f = func_1, data = h)

    # Check that it's set up properly
    assert chi_squared.func_code.co_varnames == ["a", "b"]

    # Calculate the chi_squared for the given parameters.
    result = chi_squared(np.array(range(-1, -6, -1)), np.zeros(5))
    # Each term is (1 - -1)^2 / 1^2 = 4
    assert result == 4 * 5

#####################################
# Testing cost functions against ROOT
#####################################
def parabola(x: float, scale: float) -> float:
    """ Parabolic function.

    Note:
        It returns a float, but numba can't handle cast. So we return ``Any`` and then cast the result.

    Args:
        x: Where the parabola will be evaluated.
        scale: Scale factor for the parabola.
    Returns:
        Value of parabola for given parameters.
    """
    return scale * np.square(x)  # type: ignore

@pytest.fixture  # type: ignore
def setup_parabola(logging_mixin: Any) -> Tuple[histogram.Histogram1D, Hist]:
    """ Setup a parabola for tests of fitting procedures. """
    ROOT = pytest.importorskip("ROOT")

    # Specify a seed so the test is reproducible.
    np.random.seed(12345)

    h_ROOT = ROOT.TH1F("test", "test", 42, -10.5, 10.5)
    h_ROOT.Sumw2()
    for x in np.linspace(-10.25, 10.25, 42):
        # Ensure that the bin at 0 is not precisely 0
        if x == 0.0:
            h_ROOT.Fill(x, 2)

        # Adds a gaussian noise term with a width of 3. It's offset from 0 to ensure that we don't get 0.
        #for _ in np.arange(int(parabola(np.abs(x), 1) + np.random.normal(5, 4))):
        for _ in np.arange(int(np.ceil(parabola(np.abs(x), 1) + np.random.normal(3, 3)))):
            #logger.debug(f"Filling for x: {x}")
            h_ROOT.Fill(x)

    # Scale by bin width
    h_ROOT.Scale(1.0 / h_ROOT.GetBinWidth(1))

    # Convert
    h = histogram.Histogram1D.from_existing_hist(h_ROOT)
    logger.debug(f"h: {h}")

    #c = ROOT.TCanvas("c", "c")
    #h_ROOT.Draw()
    #c.SaveAs("test_parabola.pdf")

    return h, h_ROOT

@pytest.mark.parametrize("cost_func, fit_option", [  # type: ignore
    (cost_function.BinnedChiSquared, "SV"),
    (cost_function.BinnedLogLikelihood, "SLV"),
    (cost_function.BinnedLogLikelihood, "SWLV"),
    ("probfit", "SV"),
], ids = ["Binned chi squared", "Binned log likelihood", "Binned log likelihood with weighting", "Probfit Chi2"])
def test_binned_cost_functions_against_ROOT(logging_mixin: Any, cost_func: Any, fit_option: Any,
                                            setup_parabola: Any) -> None:
    """ Test the binned cost function implementations against ROOT. """
    # Setup
    h, h_ROOT = setup_parabola
    ROOT = pytest.importorskip("ROOT")
    minuit_args: Dict[str, Union[float, Tuple[float, float]]] = {
        "scale": 1, "error_scale": 0.1, "limit_scale": (-1000, 1000),
    }
    log_likelihood = "L" in fit_option
    if cost_func == "probfit":
        probfit = pytest.importorskip("probfit")
        cost_func = probfit.Chi2Regression

    # Fit with ROOT
    fit_ROOT = ROOT.TF1("parabola", "[0] * TMath::Power(x, 2)", -10.5, 10.5)
    # Expect it to be around 1.
    fit_ROOT.SetParameter(0, minuit_args["scale"])
    fit_result_ROOT = h_ROOT.Fit(fit_ROOT, fit_option + "0")
    logger.debug(f"ROOT: chi_2: {fit_result_ROOT.Chi2()}, ndf: {fit_result_ROOT.Ndf()}")

    # Fit with the defined cost function
    args: Dict[str, Any] = {"f": parabola}
    if issubclass(cost_func, cost_function.CostFunctionBase):
        args.update({"data": h})
        # Test for weighted likelihood
        if "W" in fit_option:
            args.update({"use_weights": True})
    else:
        args.update({"x": h.x, "y": h.y, "error": h.errors})
    cost = cost_func(**args)
    fit_result, minuit = fit_integration.fit_with_minuit(cost, minuit_args, h.x)

    # Check the minimized value.
    # There is still something a bit different between ROOT's log likelihood calculation and mine.
    # However, the other parameters appear to agree, so it seems okay.
    if not log_likelihood:
        assert np.isclose(fit_result.minimum_val, fit_result_ROOT.MinFcnValue(), rtol = 0.03)

    if cost_func is cost_function.BinnedLogLikelihood:
        # Calculate the chi squared equivalent and set that to be the minimum value for comparison.
        binned_chi_squared = cost_function._binned_chi_squared(
            h.x, h.y, h.errors, h.bin_edges, parabola, *list(fit_result.values_at_minimum.values())
        )
        unbinned_chi_squared = cost_function._chi_squared(
            h.x, h.y, h.errors, h.bin_edges, parabola, *list(fit_result.values_at_minimum.values())
        )
        logger.debug(
            f"minimal_val before changing: {fit_result.minimum_val}, ROOT func min: {fit_result_ROOT.MinFcnValue()}"
        )
        logger.debug(f"binned chi_squared: {binned_chi_squared}, unbinned chi_squared: {unbinned_chi_squared}")
        fit_result.minimum_val = binned_chi_squared

    # Calculate errors.
    fit_result.errors = fit_base.calculate_function_errors(
        func = parabola,
        fit_result = fit_result,
        x = fit_result.x
    )

    # Check the result
    logger.debug(f"Fit chi_2: {fit_result.minimum_val}, ndf: {fit_result.nDOF}")
    # It won't agree exactly because ROOT appears to use the unbinned chi squared to calculate this value.
    # This can be seen because probfit agrees with ROOT.
    assert np.isclose(fit_result.minimum_val, fit_result_ROOT.Chi2(), rtol = 0.035)
    assert np.isclose(fit_result.nDOF, fit_result_ROOT.Ndf())
    # Check the parameters
    # Value
    assert np.isclose(
        fit_result.values_at_minimum["scale"], fit_result_ROOT.Parameter(0), rtol = 0.05,
    )
    # Error
    assert np.isclose(fit_result.errors_on_parameters["scale"], fit_result_ROOT.ParError(0), rtol = 0.005)
    # Covariance matrix
    if issubclass(cost_func, cost_function.CostFunctionBase):
        covariance_ROOT = fit_result_ROOT.GetCovarianceMatrix()
        # Print the fit result, alongside the covariance
        fit_result_ROOT.Print("V")
        logger.debug(f"Covariance: {fit_result.covariance_matrix}")
        for i_name in fit_result.free_parameters:
            for j_name in fit_result.free_parameters:
                i_index = fit_result.free_parameters.index(i_name)
                j_index = fit_result.free_parameters.index(j_name)
                logger.debug(f"Checking covariance matrix parameters: ({i_name}:{i_index}, {j_name}:{j_index})")
                assert np.isclose(fit_result.covariance_matrix[(i_name, j_name)], covariance_ROOT(i_index, j_index), rtol = 0.01)
    # Estimated distance to minimum
    assert np.isclose(minuit.edm, fit_result_ROOT.Edm(), atol = 1e-3)

    # Check the effective chi squared. This won't work in the probfit case because we don't recognize
    # the type properly (and it's not worth the effort).
    if issubclass(cost_func, cost_function.CostFunctionBase):
        assert fit_result.effective_chi_squared(cost) == (
            cost_function._binned_chi_squared(
                cost.data.x, cost.data.y, cost.data.errors, cost.data.bin_edges,
                cost.f, *fit_result.values_at_minimum.values()
            ) if log_likelihood else fit_result.minimum_val
        )


##################
# Simultaneous Fit
##################

@pytest.fixture  # type: ignore
def setup_simultaneous_fit_data(logging_mixin: Any, setup_parabola: Any) -> Tuple[histogram.Histogram1D, histogram.Histogram1D, Hist, Hist]:
    """ Setup the data for tests of a simultaneous fit. """
    h, h_ROOT = setup_parabola

    # Create a new parabola that's shifted up by two.
    h_shifted_ROOT = h_ROOT.Clone("shifted_parabola")
    h_shifted_ROOT.Add(h_ROOT)

    h_shifted = histogram.Histogram1D.from_existing_hist(h_shifted_ROOT)

    return h, h_shifted, h_ROOT, h_shifted_ROOT

def test_simultaneous_fit_basic(logging_mixin: Any, setup_simultaneous_fit_data: Any) -> None:
    """ Test basic Simultaneous fit functionality. """
    # Setup
    h, h_shifted, _, _ = setup_simultaneous_fit_data

    # Check with cost functions
    cost_func1 = cost_function.ChiSquared(func_1, data = h)
    cost_func2 = cost_function.ChiSquared(func_2, data = h_shifted)
    s2 = cost_function.SimultaneousFit(cost_func1, cost_func2)
    assert s2.func_code == fit_base.FuncCode(["a", "b", "c", "d"])

    # Check with manually added functions
    s3 = cost_func1 + cost_func2
    assert s3.func_code == fit_base.FuncCode(["a", "b", "c", "d"])
    assert s3 == s2

def test_nested_simultaneous_fit_objects(logging_mixin: Any, setup_simultaneous_fit_data: Any) -> None:
    """ Test for unraveling nested simultaneous fit objects. """
    # Setup
    h, h_shifted, _, _ = setup_simultaneous_fit_data

    # Check with cost functions
    cost_func1 = cost_function.ChiSquared(func_1, data = h)
    cost_func2 = cost_function.ChiSquared(func_2, data = h_shifted)
    cost_func3 = cost_function.ChiSquared(lambda x, e, f: x + e + f, data = h)
    s = cost_func1 + cost_func2
    s2 = s + cost_func3
    assert s2.func_code == fit_base.FuncCode(["a", "b", "c", "d", "e", "f"])

    # Test out using sum
    s3 = sum([cost_func1, cost_func2, cost_func3])
    # Help out mypy...
    assert isinstance(s3, cost_function.SimultaneousFit)
    assert s3.func_code == fit_base.FuncCode(["a", "b", "c", "d", "e", "f"])

def test_simultaneous_fit(logging_mixin: Any, setup_simultaneous_fit_data: Any) -> None:
    """ Test Simultaneous Fit functionality vs probfit with an integration test. """
    # Setup
    h, h_shifted, _, _ = setup_simultaneous_fit_data
    cost_func1 = cost_function.ChiSquared(parabola, data = h)
    cost_func2 = cost_function.ChiSquared(parabola, data = h_shifted)
    minuit_args: Dict[str, Union[float, Tuple[float, float]]] = {
        "scale": 1.5, "error_scale": 0.15, "limit_scale": (-1000, 1000),
    }

    # Setup the probfit version
    probfit = pytest.importorskip("probfit")
    s_probfit = probfit.SimultaneousFit(*[cost_func1, cost_func2])

    # Setup the comparison version
    s = cost_func1 + cost_func2

    # First, basic checks
    logger.debug(f"func_code: {s.func_code}, co_varnames: {s.func_code.co_varnames}")
    assert s.func_code == fit_base.FuncCode(["scale"])
    assert s.func_code.co_varnames == list(s_probfit.func_code.co_varnames)

    # Now perform the fits
    fit_result, _ = fit_integration.fit_with_minuit(
        cost_func = s, minuit_args = minuit_args, x = h.x
    )
    fit_result_probfit, _ = fit_integration.fit_with_minuit(
        cost_func = s_probfit, minuit_args = minuit_args, x = h.x
    )
    # And check that the fit results agree
    logger.debug(f"scale: {fit_result.values_at_minimum['scale']} +/- {fit_result.errors_on_parameters['scale']}")
    assert fit_result == fit_result_probfit



#!/usr/bin/env python3

""" Models for fitting.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import abc
import logging
import operator
from typing import Any, Callable, Dict, Iterable, Iterator, TypeVar, Union

import iminuit
import numpy as np
import scipy.integrate

from pachyderm import generic_class, histogram
from pachyderm.fit import base as fit_base

logger = logging.getLogger(__name__)

T_CostFunction = TypeVar("T_CostFunction", bound = "CostFunctionBase")

def _quad(f: Callable[..., float], bin_edges: np.ndarray, *args: Union[float, np.ndarray]) -> np.ndarray:
    """ Integrate over the given function using QUADPACK.

    Unfortunately, this option is fairly slow because we can't take advantage of vectorized numpy operations.
    Something like numba could speed this up if all functions and classes could be supported.

    Args:
        f: Function to integrate.
        bin_edges: Bin edges of the data histogram.
        args: Arguments for evaluating the function.
    Returns:
        Integral over each bin.
    """
    values = []
    for lower, upper in zip(bin_edges[:-1], bin_edges[1:]):
        res, _ = scipy.integrate.quad(func = f, a = lower, b = upper, args = tuple(args))
        values.append(res)
    return np.array(values)

def _simpson_38(f: Callable[..., float], bin_edges: np.ndarray, *args: Union[float, np.ndarray]) -> np.ndarray:
    """ Integrate over each histogram bin with the Simpson 3/8 rule.

    Implemented via the expression at https://en.wikipedia.org/wiki/Simpson%27s_rule#Simpson's_3/8_rule .

    Could also be implemented via ``scipy.integrate.simps``, but it implements the composite
    rule which, as far as I can tell, doesn't map onto our histogram data quite as nicely
    as our own implementation.

    Args:
        f: Function to integrate.
        bin_edges: Bin edges of the data histogram.
        args: Arguments for evaluating the function.
    Returns:
        Integral over each bin.
    """
    a = bin_edges[:-1]
    b = bin_edges[1:]
    # Recall that bin_edges[1:] - bin_edges[:-1] is the bin widths
    return (b - a) / 8 * (f(a, *args) + 3 * f((2 * a + b) / 3, *args) + 3 * f((a + 2 * b) / 3, *args) + f(b, *args))

def _integrate_1D(f: Callable[..., float], bin_edges: np.ndarray, *args: Union[float, np.ndarray]) -> np.ndarray:
    """ Integrate the given function over each bin in 1D.

    A number of options are available for integration, including a simple method evaluated on
    the bin edges, Simpson's 3/8 rule, and using QUADPACK.

    Note:
        Integration depends on the bin edges, which implicitly undoes the bin width scaling that
        is applied to a histogram. To compensate, we divide the integrated values by the bin width.

    Args:
        f: Function to integrate.
        bin_edges: Bin edges of the data histogram.
        args: Arguments for evaluating the function.
    Returns:
        Integral over each bin.
    """
    # Simplest case where we just evaluate on the edges.
    #return (f(bin_edges[1:], *args) + f(bin_edges[:-1], *args)) / 2 * (bin_edges[1:] - bin_edges[:-1])
    # QUADPACK is another option, but it's slow.
    #return _quad(f, bin_edges, *args) / (bin_edges[1:] - bin_edges[:-1])
    # Simpson's 3/8 rule is better than the simple case, but faster than QUADPACK.
    return _simpson_38(f, bin_edges, *args) / (bin_edges[1:] - bin_edges[:-1])

def unravel_simultaneous_fits(functions: Iterable[Union["CostFunctionBase", "SimultaneousFit"]]
                              ) -> Iterator["CostFunctionBase"]:
    """ Unravel the cost functions from possible simultaneous fit objects.

    The functions are unravel by recursively retrieving the functions from existing ``SimultaneousFit`` objects
    that may be in the list of passed functions. The cost functions store their fit data, so they are fully
    self contained. Consequently, we are okay to fully unravel the functions without worrying about the
    intermediate ``SimultaneousFit`` objects.

    Args:
        functions: Functions to unravel.
    Returns:
        Iterator of the base cost functions.
    """
    for f in functions:
        if isinstance(f, SimultaneousFit):
            yield from unravel_simultaneous_fits(f.cost_functions)
        else:
            yield f

class SimultaneousFit(generic_class.EqualityMixin):
    """ Cost function for the simultaneous fit of the given cost functions.

    Args:
        cost_functions: The cost functions.

    Attributes:
        functions: The cost functions.
        func_code: Function arguments derived from the fit functions. They need to be separately
            specified to allow iminuit to determine the proper arguments.
        argument_positions: Map of merged arguments to the arguments for each individual function.
    """
    def __init__(self, *cost_functions: Union[T_CostFunction, "SimultaneousFit"]):
        # Validation
        # Ensure that we unravel any SimultaneousFit objects to their base cost functions.
        funcs = list(unravel_simultaneous_fits(list(cost_functions)))

        self.cost_functions = funcs
        logger.debug("Simultaneous Fit")
        merged_args, argument_positions = fit_base.merge_func_codes(self.cost_functions)
        # We don't drop any of the arguments here because the cost functions already did it.
        self.func_code = fit_base.FuncCode(merged_args)
        self.argument_positions = argument_positions

    def __add__(self, other: Union[T_CostFunction, "SimultaneousFit"]) -> "SimultaneousFit":
        """ Add a new function to the simultaneous fit. """
        return type(self)(self, other)

    def __radd__(self, other: Union[T_CostFunction, "SimultaneousFit"]) -> "SimultaneousFit":
        """ For use with ``sum(...)``. """
        if other == 0:
            return self
        else:
            return self + other

    def __call__(self, *args: float) -> float:
        """ Calculate the cost function for all x values in the data. """
        return fit_base.call_list_of_callables_with_operation(
            operator.add, self.cost_functions, self.argument_positions, *args
        )

class CostFunctionBase(abc.ABC):
    """ Base cost function.

    Args:
        f: The fit function.
        data: Data to be used for fitting.
        additional_call_options: Additional keyword options to be passed when calling the cost function.
    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Data to be used for fitting.
        additional_call_options: Additional keyword options to be passed when calling the cost function.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    _cost_function: Callable[..., float]

    def __init__(self, f: Callable[..., float], data: Union[np.ndarray, histogram.Histogram1D],
                 **additional_call_options: Any):
        # If using numba, we would need to JIT the function to be able to pass it to the cost function.
        self.f = f
        # We need to drop the leading x argument
        self.func_code = fit_base.FuncCode(iminuit.util.describe(self.f)[1:])
        self.data = data
        self._additional_call_options: Dict[str, Any] = additional_call_options

    def __add__(self: T_CostFunction, other: T_CostFunction) -> SimultaneousFit:
        """ Creates a simultaneous fit when added with another cost function. """
        return SimultaneousFit(self, other)

    def __radd__(self: T_CostFunction, other: T_CostFunction) -> Union[T_CostFunction, SimultaneousFit]:
        """ For use with ``sum(...)``. """
        if other == 0:
            return self
        else:
            return self + other

    def __call__(self, *args: float) -> float:
        """ Calculate the cost function for all x values in the data. """
        return self._call_cost_function(self.data, self.f, *args, **self._additional_call_options)

    @classmethod
    @abc.abstractmethod
    def _call_cost_function(cls, data: Union[np.ndarray, histogram.Histogram1D],
                            f: Callable[..., float], *args: Union[float, np.ndarray], **kwargs: Any) -> float:
        """ Wrapper to allow access to the method as if it's unbound.

        This is needed for use with numba.

        Args:
            data: The input data.
            f: Fit function.
            args: Other arguments for the fit function (not including where it will be evaluated (ie. x)).
            kwargs: Additional arguments to pass to the cost function.
        Returns:
            The cost function evaluated at all of the corresponding data points.
        """
        ...

class StandaloneCostFunction(CostFunctionBase):
    """ Cost function which only needs a list of input data.

    This is in contrast to those which need data to compare against at each point. One example of
    a cost function which only needs the input data is the unbinned log likelihood.

    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Numpy array of all input values (not binned in any way). It's just a list of the values.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        # Check that we have the proper input data. This isn't very pythnoic, but it's
        # important that have the data in the proper format.
        assert isinstance(self.data, np.ndarray)

    @classmethod
    def _call_cost_function(cls, data: np.ndarray,
                            f: Callable[..., float], *args: Union[float, np.ndarray], **kwargs: Any) -> float:
        """ Wrapper to allow access to the method as if it's unbound.

        This is needed for use with numba.

        Args:
            data: The input histogram. This should simply be an array of the values.
        Returns:
            The cost function evaluated at all of the corresponding data points (ie. ``self.data``).
        """
        return cls._cost_function(data, f, *args, **kwargs)

class DataComparisonCostFunction(CostFunctionBase):
    """ Cost function which needs comparison data, the points where it was evaluated, and the errors.

    This is in contrast to those which only need the input data. Examples of cost functions needing
    input data included the chi squared (both unbinned and binned), as well as the binned log likelihood.

    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Numpy array of all input values (not binned in any way). It's just a list of the values.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        # Check that we have the proper input data. This isn't very pythnoic, but it's
        # important that have the data in the proper format.
        assert isinstance(self.data, histogram.Histogram1D)

    @classmethod
    def _call_cost_function(cls, data: Any,
                            f: Callable[..., float], *args: Union[float, np.ndarray], **kwargs: Any) -> float:
        """ Wrapper to allow access to the method as if it's unbound.

        This is needed for use with numba.

        Args:
            data: Input data in the form of a histogram. Note that this must be a ``Histogram1D``
                (but it's specified as Any above to avoid needing isinstance calls in performance
                sensitive code).
            f: Fit function.
            args: Arguments for the function.
        Returns:
            The cost function evaluated at all the corresponding data points (ie. data.x).
        """
        return cls._cost_function(data.x, data.y, data.errors, data.bin_edges, f, *args, **kwargs)

def _chi_squared(x: np.ndarray, y: np.ndarray,
                 errors: np.ndarray, _: np.ndarray,
                 f: Callable[..., float], *args: float) -> Any:
    r""" Actual implementation of the chi squared.

    Implemented with some help from the iminuit advanced tutorial. The chi squared is defined as:

    .. math::

        \Chi^{2} = \sum_{i} (\frac{(y_{i} - f(x, *args)}{error_{i})})^{2}

    Note:
        It returns a float, but numba can't handle cast. So we return ``Any`` and then cast the result.

    Args:
        x: x values for calculation.
        y: y values for calculation.
        errors: Errors for calculation.
        _: Ignored here.
        f: Fit function.
        args: Additional arguments for the fit function.
    Returns:
        (Unbinned) chi squared for the given arguments.
    """
    return np.sum(np.square((y - f(x, *args)) / errors))

class ChiSquared(DataComparisonCostFunction):
    """ chi^2 cost function.

    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Data to be used for fitting.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    _cost_function = _chi_squared

def _binned_chi_squared(x: np.ndarray, y: np.ndarray,
                        errors: np.ndarray, bin_edges: np.ndarray,
                        f: Callable[..., float], *args: float) -> Any:
    r""" Actual implementation of the binned chi squared.

    The binned chi squared is defined as:

    .. math::

        \Chi^{2} = \sum_{i} (\frac{(y_{i} - \int_{i \mathrm{lower edge}}^{i \mathrm{upper edge}} f(x, *args)}{error_{i})})^{2}

    where the function is integrated over each bin.

    Note:
        It returns a float, but numba can't handle cast. So we return ``Any`` and then cast the result.

    Args:
        x: x values where the function should be evaluated.
        y: Histogram values at each x.
        errors: Histogram errors at each x.
        bin_edges: Histogram bin edges.
        f: Fit function.
        args: Arguments for the fit function.
    Returns:
        Binned chi squared calculated for each x value.
    """
    # ROOT appears to use the unbinned chi squared, despite working with binned data
    #return np.sum(np.square((y - f(x, *args)) / errors))
    # Evaluate the function value over the entire bin.
    expected_values = _integrate_1D(f, bin_edges, *args)
    return np.sum(np.square((y - expected_values) / errors))

def binned_chi_squared_safe_for_zeros(x: np.ndarray, y: np.ndarray,
                                      errors: np.ndarray, bin_edges: np.ndarray,
                                      f: Callable[..., float], *args: float) -> Any:
    """ Actual implementation of the binned chi squared.

    See `_binned_chi_squared` for further information. This function is just the standard binned chi squared,
    but the division is protected from divide by 0. This allows safe use when calculating a binned chi squared.

    Args:
        x: x values where the function should be evaluated.
        y: Histogram values at each x.
        errors: Histogram errors at each x.
        bin_edges: Histogram bin edges.
        f: Fit function.
        args: Arguments for the fit function.
    Returns:
        Binned chi squared calculated for each x value.
    """
    expected_values = _integrate_1D(f, bin_edges, *args)
    return np.sum(np.square(
        np.divide((y - expected_values), errors, out = np.zeros_like(errors), where = errors != 0)
    ))

class BinnedChiSquared(DataComparisonCostFunction):
    """ Binned chi^2 cost function.

    Calling this class will calculate the chi squared. Implemented with some help from ...

    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Data to be used for fitting.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    _cost_function = _binned_chi_squared

def _log_likelihood(data: np.ndarray, f: Callable[..., float], *args: float) -> Any:
    r""" Actual implementation of the log likelihood cost function.

    The unbinned log likelihood is defined as:

    .. math::

        \mathrm{likelihood} = - \sum_{i \in data} \log{f(x, *args)}

    Note:
        It returns a float, but numba can't handle cast. So we return ``Any`` and then cast the result.

    Args:
        data: Data points (raw data, not histogramed).
        f: Fit function.
        args: Fit function arguments.
    Returns:
        Unbinned log likelihood for the given data points.
    """
    return np.log(f(data, *args))

class LogLikelihood(StandaloneCostFunction):
    """ Log likelihood cost function.

    Calling this class will calculate the chi squared. Implemented with some help from ...

    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Data to be used for fitting.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    _cost_function = _log_likelihood

def _extended_binned_log_likelihood(x: np.ndarray, y: np.ndarray,
                                    errors: np.ndarray, bin_edges: np.ndarray,
                                    f: Callable[..., float], *args: float, use_weights: bool = False) -> Any:
    r""" Actual implementation of the extended binned log likelihood (cost function).

    Based on Probfit's binned log likelihood implementation. I also looked at
    ROOT's ``FitUtil::EvaluatePoissonLogL(...)`` for evaluating the binned log likelihood,
    and they appear to be consistent.

    The actual expression that is implemented is:

    .. math::

        \textrm{Likelihood} = -\sum_{i \in bins} s_i \times  \left(  h_i \times \log (\frac{E_i}{h_i}) + (h_i-E_i) \right)

    where E_i is the expected value (from the fit function), and h_i is the histogram. To make a proper
    comparison between E_i and h_i, we need to evaluate the value of the entire bin. To do so, we use:

    .. math::

        E_i = \frac{f(l_i, arg\ldots )+f(r_i, arg \ldots )}{2} \times b_i

    and s_i is unity when performing a standard fit, and:

    .. math::

        s_i = h_i / error_i^2

    when performing a weighted fit.

    Note:
        It returns a float, but numba can't handle cast. So we return ``Any`` and then cast the result.

    Args:
        x: x values for calculation.
        y: y values for calculation.
        errors: Errors for calculation.
        bin_edges: Bin widths of the histogram.
        f: Fit function.
        args: Additional arguments for the fit function.
        use_weights: Use data weights in calculating the log likelihood.
    Returns:
        Binned log likelihood.
    """
    # Need to normalize the contributions.
    scale = y / errors ** 2 if use_weights else np.ones(len(y))
    expected_values = _integrate_1D(f, bin_edges, *args)
    # We don't use log rules to combine the log1p expressions (ie. log(expected_values / y)) because it appears
    # to create numerical issues (throwing NaN).
    # It sounds like the absolute value of FCN doesn't necessarily mean much for the log likelihood ratio
    # In principle, I should be able to get it to match ROOT, but that doesn't seem so trivial in practice.
    return -1 * np.sum(scale * (y * (np.log1p(expected_values) - np.log1p(y)) + (y - expected_values)))

class BinnedLogLikelihood(DataComparisonCostFunction):
    """ Binned log likelihood cost function.

    Calling this class will calculate the chi squared. Implemented with some help from ...

    Args:
        f: The fit function.
        data: Data to be used for fitting.
        use_weights: Whether to use the data weights when calculating the cost function.
            This is equivalent to the "WL" option in ROOT. Default: False.
        additional_call_options: Additional keyword options to be passed when calling the cost function.
    Attributes:
        f: The fit function.
        func_code: Function arguments derived from the fit function. They need to be separately specified
            to allow iminuit to determine the proper arguments.
        data: Data to be used for fitting.
        _cost_function: Function to be used to calculate the actual cost function.
    """
    def __init__(self, use_weights: bool = False, *args: Any, **kwargs: Any):
        kwargs["use_weights"] = use_weights
        super().__init__(*args, **kwargs)

    _cost_function = _extended_binned_log_likelihood



#!/usr/bin/env python

""" Module related to YAML.

Contains a way to construct the main YAML object, as well as relevant mixins and classes.

Note:
    The YAML to/from enum values would be much better as a mixin. However, such an approach causes substantial issues.
    In particular, although we don't explicitly pickle the values, calling ``copy.copy`` implicitly calls pickle, so
    we must maintain compatibility. However, enum mixins preclude pickling the enum value
    (see `cpython/enum.py line 177 <https://github.com/python/cpython/blob/master/Lib/enum.py#L177>`__). The problem
    basically comes down to the fact that we are assigning a bound staticmethod to the class when we mix
    it in, and it doesn't seem to be able to resolving pickling the object (perhaps due to name resolution issues).
    For a bit more, see the comments `on this stackoverflow post <https://stackoverflow.com/q/46230799>`__.
    Piratically, I believe that we could also resolve this by implementing ``__reduce_ex``, but that appears as if
    it will be more work than our implemented workaround. Our workaround can be implemented as:

    .. code-block:: python

        >>> class TestEnum(enum.Enum):
        ...   a = 1
        ...   b = 2
        ...
        ...   def __str__(self):
        ...     return self.name
        ...
        ...   to_yaml = staticmethod(generic_class.enum_to_yaml)
        ...   from_yaml = staticmethod(generic_class.enum_from_yaml)

    This enum object will pickle properly. Note that rather strangely, this issue showed up during tests on Debian
    Stretch, but not the exact same version of python on macOS. I don't know why that's the case, but the workaround
    seems to be fine on both systems, so we'll just continue to use it.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import base64
import enum
import inspect
import logging
from io import BytesIO
from typing import Any, Iterable, Optional, Type, TypeVar, cast

import numpy as np
import ruamel.yaml

logger = logging.getLogger(__name__)

# Typing helpers
DictLike = ruamel.yaml.comments.CommentedMap
Representer = Type[ruamel.yaml.representer.BaseRepresenter]
Constructor = Type[ruamel.yaml.constructor.BaseConstructor]
T_EnumToYAML = TypeVar("T_EnumToYAML", bound = enum.Enum)
T_EnumFromYAML = TypeVar("T_EnumFromYAML", bound = enum.Enum)

def yaml(modules_to_register: Optional[Iterable[Any]] = None,
         classes_to_register: Optional[Iterable[Any]] = None) -> ruamel.yaml.YAML:
    """ Create a YAML object for loading a YAML configuration.

    Args:
        modules_to_register: Modules containing classes to be registered with the YAML object. Default: None.
        classes_to_register: Classes to be registered with the YAML object. Default: None.
    Returns:
        A newly creating YAML object, configured as appropriate.
    """
    # Define a round-trip YAML object for us to work with. This object should be imported by other modules
    # NOTE: "typ" is a not a typo. It stands for "type"
    yaml = ruamel.yaml.YAML(typ = "rt")

    # Register representers and constructors
    # Numpy array
    yaml.representer.add_representer(np.ndarray, numpy_array_to_yaml)
    yaml.constructor.add_constructor("!numpy_array", numpy_array_from_yaml)
    # Numpy float64
    yaml.representer.add_representer(np.float64, numpy_float64_to_yaml)
    yaml.constructor.add_constructor("!numpy_float64", numpy_float64_from_yaml)
    # Register external classes
    yaml = register_module_classes(yaml = yaml, modules = modules_to_register)
    yaml = register_classes(yaml = yaml, classes = classes_to_register)

    return yaml

def register_classes(yaml: ruamel.yaml.YAML, classes: Optional[Iterable[Any]] = None) -> ruamel.yaml.YAML:
    """ Register externally defined classes. """
    # Validation
    if classes is None:
        classes = []

    # Register the classes
    for cls in classes:
        logger.debug(f"Registering class {cls} with YAML")
        yaml.register_class(cls)

    return yaml

def register_module_classes(yaml: ruamel.yaml.YAML, modules: Optional[Iterable[Any]] = None) -> ruamel.yaml.YAML:
    """ Register all classes in the given modules with the YAML object.

    This is a simple helper function.
    """
    # Validation
    if modules is None:
        modules = []

    # Extract the classes from the modules
    classes_to_register = set()
    for module in modules:
        module_classes = [member[1] for member in inspect.getmembers(module, inspect.isclass)]
        classes_to_register.update(module_classes)

    # Register the extracted classes
    return register_classes(yaml = yaml, classes = classes_to_register)

#
# Representers and constructors for individual classes.
#

def numpy_array_to_yaml(representer: Representer, data: np.ndarray) -> str:
    """ Write a numpy array to YAML.

    It registers the array under the tag ``!numpy_array``.

    Use with:

    .. code-block:: python

        >>> yaml = ruamel.yaml.YAML()
        >>> yaml.representer.add_representer(np.ndarray, yaml.numpy_array_to_yaml)

    Note:
        We cannot use ``yaml.register_class`` because it won't register the proper type.
        (It would register the type of the class, rather than of `numpy.ndarray`). Instead,
        we use the above approach to register this method explicitly with the representer.
    """
    # Create a bytes object, dump to it, encode the bytes to a str, and then write them.
    # It's less transparent when physically reading it, but it should avoid encoding issues.
    b = BytesIO()
    np.save(b, data)
    b.seek(0)
    # The representer is seen by mypy as Any, so we need to explicitly note that it's a str.
    return cast(
        str,
        representer.represent_scalar(
            "!numpy_array", base64.encodebytes(b.read()).decode("utf-8"),
        )
    )

def numpy_array_from_yaml(constructor: Constructor, data: ruamel.yaml.nodes.SequenceNode) -> np.ndarray:
    """ Read an array from YAML to numpy.

    It reads arrays registered under the tag ``!numpy_array``.

    Use with:

    .. code-block:: python

        >>> yaml = ruamel.yaml.YAML()
        >>> yaml.constructor.add_constructor("!numpy_array", yaml.numpy_array_from_yaml)

    Note:
        We cannot use ``yaml.register_class`` because it won't register the proper type.
        (It would register the type of the class, rather than of `numpy.ndarray`). Instead,
        we use the above approach to register this method explicitly with the representer.

    Note:
        In order to allow users to write an array by hand, we check the data given. If it's
        a list, we convert the values and put them into an array. If it's binary encoded,
        we decode and load it.

    Args:
        constructor: YAML constructor being used to read and create the objects specified in the YAML.
        data: Data stored in the YAML node currently being processed.
    Returns:
        Numpy array containing the data in the current YAML node.
    """
    return_value: np.ndarray
    if isinstance(data.value, list):
        # These are probably from a hand encoded file. We will convert them into an array.
        # Construct the contained values so that we properly construct int, float, etc.
        # We just leave this to YAML because it already stores this information.
        values = [constructor.construct_object(n) for n in data.value]
        return_value = np.array(values)
    else:
        # Binary encoded numpy. Decode and load it.
        b = data.value.encode("utf-8")
        # Requires explicitly allowing pickle to load arrays. This used to be default True,
        # so our risk hasn't changed.
        return_value = np.load(BytesIO(base64.decodebytes(b)), allow_pickle = True)
    return return_value

def numpy_float64_to_yaml(representer: Representer, data: np.float64) -> str:
    """ Write a numpy float64 to YAML.

    It registers the float under the tag ``!numpy_float64``.

    Use with:

    .. code-block:: python

        >>> yaml = ruamel.yaml.YAML()
        >>> yaml.representer.add_representer(np.float64, yaml.numpy_float64_to_yaml)

    Note:
        We cannot use ``yaml.register_class`` because it won't register the proper type.
        (It would register the type of the class, rather than of `numpy.float64`). Instead,
        we use the above approach to register this method explicitly with the representer.
    """
    # Create a bytes object, dump to it, encode the bytes to a str, and then write them.
    # It's less transparent when physically reading it, but it should avoid encoding issues.
    b = BytesIO()
    np.save(b, data)
    b.seek(0)
    # The representer is seen by mypy as Any, so we need to explicitly note that it's a str.
    return cast(
        str,
        representer.represent_scalar(
            "!numpy_float64", base64.encodebytes(b.read()).decode("utf-8"),
        )
    )

def numpy_float64_from_yaml(constructor: Constructor, data: ruamel.yaml.nodes.ScalarNode) -> np.float64:
    """ Read an float64 from YAML to numpy.

    It reads the float64 registered under the tag ``!numpy_float64``.

    Use with:

    .. code-block:: python

        >>> yaml = ruamel.yaml.YAML()
        >>> yaml.constructor.add_constructor("!numpy_float64", yaml.numpy_float64_from_yaml)

    Note:
        We cannot use ``yaml.register_class`` because it won't register the proper type.
        (It would register the type of the class, rather than of `numpy.float64`). Instead,
        we use the above approach to register this method explicitly with the representer.

    Note:
        In order to allow users to write an float by hand, we check the data given. If it's
        a raw float, we put it into an float64. If it's binary encoded, we decode and load it.

    Args:
        constructor: YAML constructor being used to read and create the objects specified in the YAML.
        data: Data stored in the YAML node currently being processed.
    Returns:
        Numpy float64 containing the data in the current YAML node.
    """
    return_value: np.float64
    try:
        # First guess that it's a hand encoded file. We can't detect the type because the value
        # is just a str. We will convert it into a numpy float.
        return_value = np.float64(data.value)
    except ValueError:
        # It can't convert to a float, so it's probably binary encoded. Decode and load it.
        b = data.value.encode("utf-8")
        # Requires explicitly allowing pickle to load arrays. This used to be default True,
        # so our risk hasn't changed.
        return_value = np.load(BytesIO(base64.decodebytes(b)), allow_pickle = True)
    return return_value

def enum_to_yaml(cls: Type[T_EnumToYAML],
                 representer: Representer, data: T_EnumToYAML) -> ruamel.yaml.nodes.ScalarNode:
    """ Encodes YAML representation.

    This is a mixin method for writing enum values to YAML. It needs to be added to the enum
    as a classmethod. See the module docstring for further information on this approach and how
    to implement it.

    This method writes whatever is used in the string representation of the YAML value.
    Usually, this will be the unique name of the enumeration value. If the name is used,
    the corresponding ``EnumFromYAML`` mixin can be used to recreate the value. If the name
    isn't used, more care may be necessary, so a ``from_yaml`` method for that particular
    enumeration may be necessary.

    Note:
        This method assumes that the name of the enumeration value should be stored as a scalar node.

    Args:
        representer: Representation from YAML.
        data: Enumeration value to be encoded.
    Returns:
        Scalar representation of the name of the enumeration value.
    """
    return representer.represent_scalar(
        f"!{cls.__name__}",
        f"{str(data)}"
    )

def enum_from_yaml(cls: Type[T_EnumFromYAML],
                   constructor: Constructor, node: ruamel.yaml.nodes.ScalarNode) -> T_EnumFromYAML:
    """ Decode YAML representation.

    This is a mixin method for reading enum values from YAML. It needs to be added to the enum
    as a classmethod. See the module docstring for further information on this approach and how
    to implement it.

    Note:
        This method assumes that the name of the enumeration value was stored as a scalar node.

    Args:
        constructor: Constructor from the YAML object.
        node: Scalar node extracted from the YAML being read.
    Returns:
        The constructed YAML value from the name of the enumerated value.
    """
    return cls[node.value]



#!/usr/bin/env python3

""" Interface for fitting with Minuit.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

__all__ = [
    "base",
    "cost_function",
    "function",
    "integration",
]

from .base import (  # noqa: F401
    BaseFitResult, FitFailed, FitResult, FuncCode, calculate_function_errors, chi_squared_probability,
    evaluate_gradient, extract_function_values
)
from .cost_function import (  # noqa: F401
    BinnedChiSquared, BinnedLogLikelihood, ChiSquared, CostFunctionBase, LogLikelihood, SimultaneousFit,
    binned_chi_squared_safe_for_zeros
)
from .function import AddPDF, DividePDF, MultiplyPDF, SubtractPDF, extended_gaussian, gaussian  # noqa: F401
from .integration import Fit, T_FitArguments, fit_with_minuit  # noqa: F401



#!/usr/bin/env python

""" Objects related fixtures to aid testing.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@yale.edu>, Yale University
"""

from dataclasses import dataclass
from typing import Any, Callable, Tuple

import numpy as np
import pytest

@pytest.fixture  # type: ignore
def test_root_hists() -> Any:
    """ Create minimal TH*F hists in 1D, 2D, and 3D. Each has been filled once.

    Args:
        None
    Returns:
        tuple: (TH1F, TH2F, TH3F) for testing
    """
    import ROOT

    @dataclass
    class RootHists:
        """ ROOT histograms for testing.

        Just for convenience.
        """
        hist1D: ROOT.TH1
        hist2D: ROOT.TH2
        hist3D: ROOT.TH3

    # Define the hist to use for testing
    hist = ROOT.TH1F("test", "test", 10, 0, 1)
    hist.Fill(.1)
    hist2D = ROOT.TH2F("test2", "test2", 10, 0, 1, 10, 0, 20)
    hist2D.Fill(.1, 1)
    hist3D = ROOT.TH3F("test3", "test3", 10, 0, 1, 10, 0, 20, 10, 0, 100)
    hist3D.Fill(.1, 1, 10)

    return RootHists(hist1D = hist, hist2D = hist2D, hist3D = hist3D)

@pytest.fixture  # type: ignore
def setup_non_uniform_binning() -> Any:
    """ Test a ROOT histogram with non-uniform binning.

    Args:
        None
    Returns:
        1D histogram with non-uniform binning.
    """
    import ROOT

    binning = np.array([0, 1, 2, 4, 5, 6], dtype = np.float64)
    hist = ROOT.TH1F("test", "test", 5, binning)
    hist.Fill(1.5)

    return hist

@pytest.fixture  # type: ignore
def test_sparse() -> Any:
    """ Create a THnSparseF for testing.

    Fills in a set of values for testing.

    Args:
        None.
    Returns:
        tuple: (THnSparseF, fill_value) for testing.
    """
    import ROOT

    @dataclass
    class SparseAxis:
        """ THnSparse axis information.

        Just for convenience.
        """
        n_bins: int
        min: float
        max: float

    ignored_axis   = SparseAxis(n_bins =  1, min =   0.0, max =  1.0)  # noqa: E221, E222
    selected_axis1 = SparseAxis(n_bins = 10, min =   0.0, max = 20.0)  # noqa: E222
    selected_axis2 = SparseAxis(n_bins = 20, min = -10.0, max = 10.0)
    selected_axis3 = SparseAxis(n_bins = 30, min =   0.0, max = 30.0)  # noqa: E222
    # We want to select axes 2, 4, 5
    axes = [ignored_axis, ignored_axis, selected_axis1, ignored_axis, selected_axis2, selected_axis3, ignored_axis]

    # Create the actual sparse
    # NOTE: dtype is required here for the sparse to be created successfully.
    bins = np.array([el.n_bins for el in axes], dtype=np.int32)
    mins = np.array([el.min for el in axes])
    maxes = np.array([el.max for el in axes])
    #logger.debug("bins: {}, mins: {}, maxs: {}".format(bins, mins, maxes))
    sparse = ROOT.THnSparseF("testSparse", "testSparse", len(axes), bins, mins, maxes)

    # Fill in some strategic values.
    # Wrapper function is for convenience.
    def fill_sparse(one: float, two: float, three: float) -> None:
        # NOTE: For whatever reason, this _has_ to be float64 even though this is a
        #       SparseF. Apparently switching to a SparseD also works with float64,
        #       so something strange seems to be happening internally. But since
        #       float64 works, we stick with it.
        sparse.Fill(np.array([0., 0., one, 0., two, three, 0.], dtype = np.float64))
    fill_values = [
        (4., -2., 10.),
        (4., 2., 10.)
    ]
    for values in fill_values:
        fill_sparse(*values)

    return (sparse, fill_values)

@pytest.fixture  # type: ignore
def simple_test_functions() -> Tuple[Callable[[float, float, float], float],
                                     Callable[[float, float, float], float]]:
    """ Define simple test functions for use in tests.

    Args:
        None.
    Returns:
        func 1 (args: x, a, b), func 2 (args: x, c, d)
    """
    def func_1(x: float, a: float, b: float) -> float:
        """ Test function. """
        return x + a + b

    def func_2(x: float, c: float, d: float) -> float:
        """ Test function 2. """
        return x + c + d
    return func_1, func_2



#!/usr/bin/env python

""" Tests for the base of the fit module.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

from typing import Any, List

import iminuit
import numpy as np
import pytest

import pachyderm.fit.base as fit_base

def test_func_code(logging_mixin: Any, simple_test_functions: Any) -> None:
    """ Test creating function codes with FuncCode. """
    # Setup
    func_1, func_2 = simple_test_functions

    # Define the func code and check the properties
    func_code = fit_base.FuncCode(iminuit.util.describe(func_1))
    assert func_code.co_varnames == ["x", "a", "b"]
    assert func_code.co_argcount == 3

    # Alternatively create via static method
    func_code_2 = fit_base.FuncCode.from_function(func_1)
    assert func_code == func_code_2

@pytest.mark.parametrize("function_list_names, expected_result, expected_argument_positions", [  # type: ignore
    ([1, 2], ["x", "a", "b", "c", "d"], [[0, 1, 2], [0, 3, 4]]),
    ([2, 1], ["x", "c", "d", "a", "b"], [[0, 1, 2], [0, 3, 4]]),
], ids = ["1, 2", "2, 1"])
def test_merge_func_code(logging_mixin: Any, simple_test_functions: Any,
                         function_list_names: List[int], expected_result: List[str],
                         expected_argument_positions: List[List[int]]) -> None:
    """ Test merging function codes for a list of functions. """
    funcs = simple_test_functions
    function_list = [funcs[f_label - 1] for f_label in function_list_names]
    result, argument_positions = fit_base.merge_func_codes(function_list)
    assert result == expected_result
    assert argument_positions == expected_argument_positions

@pytest.mark.parametrize("function_list_names", [  # type: ignore
    (1, 2),
    (2, 1),
], ids = ["1, 2", "2, 1"])
def test_merge_func_code_against_probfit(logging_mixin: Any, simple_test_functions: Any, function_list_names: List[int]) -> None:
    """ Test merging function codes against probfit. """
    # Setup
    funcs = simple_test_functions
    function_list = tuple([funcs[f_label - 1] for f_label in function_list_names])
    probfit = pytest.importorskip("probfit")

    # Run the test and check
    result, argument_positions = fit_base.merge_func_codes(function_list)
    probfit_result, probfit_argument_positions = probfit.merge_func_code(*function_list)
    assert result == list(probfit_result.co_varnames)
    np.testing.assert_allclose(argument_positions, probfit_argument_positions)



#1/usr/bin/env python3

""" Base module for performing fits with Minuit.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import functools
import itertools
import logging
import time
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Optional, Sequence, Tuple, Type, TypeVar, Union, cast
)

import iminuit
import numdifftools as nd
import numpy as np

from pachyderm import generic_class

if TYPE_CHECKING:
    from pachyderm.fit import cost_function

logger = logging.getLogger(__name__)

# Typing
T_FuncCode = TypeVar("T_FuncCode", bound = "FuncCode")
T_ArgumentPositions = List[List[int]]
_T_FitResult = TypeVar("_T_FitResult", bound = "FitResult")

class FitFailed(Exception):
    """ Raised if the fit failed. The message will include further details. """
    pass

@dataclass
class BaseFitResult:
    """ Base fit result.

    This represents the most basic fit result.

    Attributes:
        parameters: Names of the parameters used in the fit.
        free_parameters: Names of the free parameters used in the fit.
        fixed_parameters: Names of the fixed parameters used in the fit.
        values_at_minimum: Contains the values of the full RP fit function at the minimum. Keys are the
            names of parameters, while values are the numerical values at convergence.
        errors_on_parameters: Contains the values of the errors associated with the parameters
            determined via the fit.
        covariance_matrix: Contains the values of the covariance matrix. Keys are tuples
            with (param_name_a, param_name_b), and the values are covariance between the specified parameters.
            Note that fixed parameters are _not_ included in this matrix.
        errors: Store the errors associated with the component fit function.
    """
    parameters: List[str]
    free_parameters: List[str]
    fixed_parameters: List[str]
    values_at_minimum: Dict[str, float]
    errors_on_parameters: Dict[str, float]
    covariance_matrix: Dict[Tuple[str, str], float]
    errors: np.ndarray

    @property
    def correlation_matrix(self) -> Dict[Tuple[str, str], float]:
        """ The correlation matrix of the free parameters.

        These values are derived from the covariance matrix values stored in the fit.

        Note:
            This property caches the correlation matrix value so we don't have to calculate it every time.

        Args:
            None
        Returns:
            The correlation matrix of the fit result.
        """
        try:
            # We attempt to cache the covariance matrix, so first try to return that.
            return self._correlation_matrix
        except AttributeError:
            def corr(i_name: str, j_name: str) -> float:
                """ Calculate the correlation matrix (definition from iminuit) from the covariance matrix. """
                # The + 1e-100 is just to ensure that we don't divide by 0.
                value = (self.covariance_matrix[(i_name, j_name)]
                         / (np.sqrt(self.covariance_matrix[(i_name, i_name)]
                            * self.covariance_matrix[(j_name, j_name)]) + 1e-100)
                         )
                # Need to explicitly cast to float. Otherwise, it will return a np.float64, which will cause problems
                # for YAML...
                return float(value)

            matrix: Dict[Tuple[str, str], float] = {}
            for i_name in self.free_parameters:
                for j_name in self.free_parameters:
                    matrix[(i_name, j_name)] = corr(i_name, j_name)

            self._correlation_matrix = matrix

        return self._correlation_matrix

@dataclass
class FitResult(BaseFitResult):
    """ Main fit result class.

    Note:
        free_parameters + fixed_parameters == parameters

    Attributes:
        parameters: Names of the parameters used in the fit.
        free_parameters: Names of the free parameters used in the fit.
        fixed_parameters: Names of the fixed parameters used in the fit.
        values_at_minimum: Contains the values of the full RP fit function at the minimum. Keys are the
            names of parameters, while values are the numerical values at convergence.
        errors_on_parameters: Contains the values of the errors associated with the parameters
            determined via the fit.
        covariance_matrix: Contains the values of the covariance matrix. Keys are tuples
            with (param_name_a, param_name_b), and the values are covariance between the specified parameters.
            Note that fixed parameters are _not_ included in this matrix.
        errors: Store the errors associated with the component fit function.
        x: x values where the fit result should be evaluated.
        n_fit_data_points: Number of data points used in the fit.
        minimum_val: Minimum value of the fit when it coverages. This is the chi squared value for a
            chi squared minimization fit.
    """
    x: np.ndarray
    n_fit_data_points: int
    minimum_val: float

    @property
    def nDOF(self) -> int:
        """ Number of degrees of freedom. """
        return self.n_fit_data_points - len(self.free_parameters)

    def effective_chi_squared(self, cost_func: "cost_function.DataComparisonCostFunction") -> float:
        """ Calculate the effective chi squared value.

        If the fit was performed using a chi squared cost function, it's just equal to
        the ``minimal_val``. If it's log likelihood, one must calculate the effective
        chi squared.

        Note:
            We attempt to cache this value so we don't have to calculate it every time.

        Args:
            cost_function: Cost function used to create the fit function.
            data: Data to be used to calculate the chi squared.
        Returns:
            The effective chi squared value.
        """
        try:
            # We attempt to cache the chi squared, so first try to return that.
            return self._chi_squared
        except AttributeError:
            # Setup
            from pachyderm.fit import cost_function
            from pachyderm import histogram

            # Calculate the chi_squared
            self._chi_squared: float
            if isinstance(cost_func, (cost_function.ChiSquared, cost_function.BinnedChiSquared)):
                self._chi_squared = self.minimum_val
            elif isinstance(cost_func, cost_function.BinnedLogLikelihood):
                data = cost_func.data
                # Help out mypy...
                assert isinstance(data, histogram.Histogram1D)
                # Calculate using the binned chi squared
                self._chi_squared = cost_function.binned_chi_squared_safe_for_zeros(
                    data.x, data.y, data.errors, data.bin_edges, cost_func.f, *self.values_at_minimum.values()
                )
            elif isinstance(cost_func, cost_function.SimultaneousFit):
                chi_squared = 0
                for cf, function_args in zip(cost_function.unravel_simultaneous_fits(cost_func.cost_functions),
                                             _function_arguments_from_argument_positions(
                                                 cost_func.argument_positions, *self.values_at_minimum.values())
                                             ):
                    data = cf.data
                    # Help out mypy...
                    assert isinstance(data, histogram.Histogram1D)
                    # Calculate using the binned chi squared. This assumes that we have binned data (which
                    # is probably fairly reasonable for our purpose).
                    chi_squared += cost_function.binned_chi_squared_safe_for_zeros(
                        data.x, data.y, data.errors, data.bin_edges, cf.f, *function_args
                    )
                self._chi_squared = chi_squared
            else:
                raise NotImplementedError("Needs to be implement for unbinned data.")
                #data = cost_func.data
                ## Help out mypy...
                #assert isinstance(data, histogram.Histogram1D)
                #self._chi_squared = cost_function._chi_squared(
                #    data.x, data.y, data.errors, data.bin_edges, cost_func.f, *self.values_at_minimum.values()
                #)

        return self._chi_squared

    @classmethod
    def from_minuit(cls: Type[_T_FitResult], minuit: iminuit.Minuit,
                    cost_func: Callable[..., float], x: np.ndarray) -> _T_FitResult:
        """ Create a fit result form the Minuit fit object.

        Args:
            minuit: Minuit fit object after performing the fit.
            cost_func: Cost function used to perform the fit.
        """
        # Validation
        if not minuit.migrad_ok():
            raise RuntimeError("The fit is invalid - unable to extract result!")

        # Determine the relevant fit parameters.
        fixed_parameters: List[str] = [k for k, v in minuit.fixed.items() if v is True]
        # We use the cost function because we want intentionally want to skip "x"
        parameters: List[str] = iminuit.util.describe(cost_func)
        # Can't just use set(parameters) - set(fixed_parameters) because set() is unordered!
        free_parameters = [p for p in parameters if p not in set(fixed_parameters)]
        # Store the result
        return cls(
            parameters = parameters, free_parameters = free_parameters, fixed_parameters = fixed_parameters,
            values_at_minimum = dict(minuit.values), errors_on_parameters = dict(minuit.errors),
            covariance_matrix = minuit.covariance,
            x = x,
            n_fit_data_points = len(x), minimum_val = minuit.fval,
            errors = [],
        )

def extract_function_values(func: Callable[..., float], fit_result: BaseFitResult) -> Tuple[Dict[str, Any], List[str]]:
    """ Extract the parameters relevant to the given function from a fit result.

    Note:
        The fit result may have more arguments at minimum and free parameters than the fit function that we've
        passed (for example, if we've calculating the background parameters for the inclusive signal fit), so
        we need to determine the free parameters here.

    Args:
        func: Function of interest.
        fit_result: Fit result from which we will extract the information.
    Returns:
        ((kw)args at minimum, free parameters)
    """
    # Setup the parameters needed to execute the function.
    # Determine relevant parameters for the given function
    func_parameters = iminuit.util.describe(func)
    # Determine the arguments for the fit function
    # NOTE: The fit result may have more arguments at minimum and free parameters than the fit function that we've
    #       passed (for example, if we've calculating the background parameters for the inclusive signal fit), so
    #       we need to determine the free parameters here.
    args_at_minimum = {k: v for k, v in fit_result.values_at_minimum.items() if k in func_parameters}
    # Retrieve the parameters to use in calculating the fit errors.
    free_parameters = [p for p in fit_result.free_parameters if p in func_parameters]

    return args_at_minimum, free_parameters

def calculate_function_errors(func: Callable[..., float], fit_result: BaseFitResult, x: np.ndarray) -> np.array:
    """ Calculate the errors of the given function based on values from the fit.

    Note:
        We don't take the x values for the fit_result as it may be desirable to calculate the errors for
        only a subset of x values. Plus, the component fit result doesn't store the x values, so it would
        complicate the validation. It's much easier to just require the user to pass the x values (and it takes
        little effort to do so).

    Args:
        func: Function to use in calculating the errors.
        fit_result: Fit result for which the errors will be calculated.
        x: x values where the errors will be evaluated.
    Returns:
        The calculated error values.
    """
    # Setup the parameters needed to execute the function.
    # Determine the arguments for the fit function and the parameters to use in calculating the fit errors.
    args_at_minimum, free_parameters = extract_function_values(func, fit_result)
    # To calculate the error, we need to match up the parameter names to their index in the arguments list
    args_at_minimum_keys = list(args_at_minimum)
    name_to_index = {name: args_at_minimum_keys.index(name) for name in free_parameters}
    logger.debug(f"args_at_minimum: {args_at_minimum}")
    logger.debug(f"free_parameters: {free_parameters}")
    logger.debug(f"name_to_index: {name_to_index}")

    # Evaluate the gradient
    partial_derivative_result = evaluate_gradient(func = func, fit_result = fit_result, x = x)

    # Finally, calculate the error by multiplying the matrix of gradient values by the covariance matrix values.
    error_vals = np.zeros(len(x))
    for i_name in free_parameters:
        for j_name in free_parameters:
            # Determine the error value
            #logger.debug(f"Calculating error for i_name: {i_name}, j_name: {j_name}")
            # Add error to overall error value
            # NOTE: This is a vector operation for the partial_derivative_result values.
            error_vals += (
                partial_derivative_result[:, name_to_index[i_name]]
                * partial_derivative_result[:, name_to_index[j_name]]
                * fit_result.covariance_matrix[(i_name, j_name)]
            )
    #logger.debug("error_val: shape: {error_val.shape}, error_val: {error_val}")

    # We want the error itself, so we take the square root.
    return np.sqrt(error_vals)

def evaluate_gradient(func: Callable[..., float], fit_result: BaseFitResult, x: np.ndarray) -> np.array:
    """ Evaluate the gradient of the given function based on the fit values.

    For a function of 5 free parameters (7 total) and 10 x values, the returned result would be of the shape (10, 5).

    Args:
        func: Function to use in calculating the errors.
        fit_result: Fit result for which the errors will be calculated.
        x: x values where the errors will be evaluated.
    Returns:
        For each x value, the gradient is evaluated for each free parameter. It will be of the
        shape (len(x_values), len(free_parameters)).
    """
    # Setup the parameters needed to execute the function.
    # Determine the arguments for the fit function and the parameters to use in calculating the fit errors.
    args_at_minimum, free_parameters = extract_function_values(func, fit_result)

    # To take the gradient, ``numdifftools`` requires a particular function signature. The first argument
    # must contain a list of values that it will vary when taking the gradient. The rest of the args are
    # passed on to the function via *args and **kwargs, but they won't be varied.
    # To ensure the proper signature, we wrap the function and route the arguments.
    def func_wrap(args_to_vary: Sequence[float], x: np.array) -> float:
        """ Wrap the given function to ensure that the arguments are routed properly for ``numdifftools``.

        To take the gradient, ``numdifftools`` requires a particular function signature. The first argument
        must contain a list of values that it will vary when taking the gradient. The rest of the args are
        passed on to the function via *args and **kwargs, but they won't be varied (we don't event use those
        generic arguments here). To ensure the proper signature given any function, we wrap the function and
        route the arguments.

        Args:
            args_to_vary: List of arguments to vary when taking the gradient. This should correspond
                to the value of the free parameters.
            x: x value(s) where the function will be evaluated.
        Returns:
            Function evaluated at the given values.
        """
        # Need to expand the arguments
        return func(x, *args_to_vary)
    # Setup to compute the derivative
    partial_derivative_func = nd.Gradient(func_wrap)

    logger.debug("Calculating the gradient")
    # We time it to keep track of how long it takes to evaluate. Sometimes it can be a bit slow.
    start = time.time()
    # Actually evaluate the gradient.
    # It returns an array with dimensions (len(x), len(args_to_vary)) containing the derivatives with
    # respect to each parameter at each point.
    #
    # NOTE: In principle, we've doing some unnecessary work because we also calculate the gradient with
    #       respect to fixed parameters. But due to the argument requirements of ``numdifftools``, it would be
    #       quite difficult to tell it to only take the gradient with respect to a non-continuous selection of
    #       parameters. So we just accept the inefficiency.
    partial_derivative_result = partial_derivative_func(list(args_at_minimum.values()), x)
    end = time.time()
    logger.debug(f"Finished calculating the gradient in {end-start} seconds.")

    # If we are only in 1D, we need to promote to a 2D (shape of 1D, 1) to get the indexing correct below.
    # This only occurs if we only have one parameter that is varied.
    if partial_derivative_result.ndim == 1:
        logger.debug(f"shape before adding axis: {partial_derivative_result.shape}")
        partial_derivative_result = partial_derivative_result[:, np.newaxis]
        logger.debug(f"shape after adding axis : {partial_derivative_result.shape}")

    return partial_derivative_result

class FuncCode(generic_class.EqualityMixin):
    """ Minimal class to describe function arguments.

    Same approach as is taken in ``iminuit``. Note that the precise name of the parameters is
    extremely important.

    Args:
        args: List of function arguments.
    Attributes:
        co_varnames: Name of the function arguments.
        co_argcount: Number of function arguments.
    """
    __slots__ = ("co_varnames", "co_argcount")

    def __init__(self, args: List[str]):
        self.co_varnames = args
        self.co_argcount = len(args)

    def __repr__(self) -> str:
        return f"FuncCode({self.co_varnames})"

    @classmethod
    def from_function(cls: Type[T_FuncCode], func: Callable[..., float],
                      leading_parameters_to_remove: int = 1) -> T_FuncCode:
        """ Create a func_code from a function.

        Args:
            func: Function for which we want a func_code.
            leading_parameters_to_remove: Number of leading parameters to remove in the func_code. Default: 1,
                which corresponds to ``x`` as the first argument.
        """
        return cls(iminuit.util.describe(func)[leading_parameters_to_remove:])

def merge_func_codes(functions: Iterable[Callable[..., float]], prefixes: Optional[Sequence[str]] = None,
                     skip_prefixes: Optional[Sequence[str]] = None) -> Tuple[List[str], List[List[int]]]:
    """ Merge the arguments of the given functions into one func_code.

    Note:
        This has very similar functionality and is heavily inspired by ``Probfit.merge_func_code...)``.

    Args:
        functions: Functions whose arguments are to be merged.
        prefixes: Prefix for arguments of each function. Default: None. If specified, there must
            be one prefix for each function.
        skip_prefixes: Prefixes to skip when assigning prefixes. As noted in probfit, this can be
            useful to mix prefixed and non-prefixed arguments. Default: None.
    Returns:
        Merged list of arguments, map from merged arguments to arguments for each individual function.
    """
    # Validation
    # Ensure that we don't exhaust the iterator during validation.
    funcs = list(functions)
    # Ensure that we have the proper number of prefixes.
    if prefixes:
        if len(funcs) != len(prefixes):
            raise ValueError("Number of prefixes ({len(prefixes)} doesn't match the number of functions: {len(funcs)}")
    else:
        # Create an empty prefix array so we can zip with it.
        prefixes = ["" for _ in funcs]
    skip_prefix = set(skip_prefixes if skip_prefixes else [])

    # Retrieve all of the args.
    args = []
    for f, pre in zip(funcs, prefixes):
        temp = []
        for arg in iminuit.util.describe(f):
            value = f"{pre}_{arg}" if pre and arg not in skip_prefix else arg
            temp.append(value)
        args.append(temp)

    # Determine the unique arguments.
    # We want to ensure that we maintain the oder, so we use dict.fromkeys to do so.
    merged_args = list(dict.fromkeys(itertools.chain.from_iterable(args)))

    # Determine the map from merged arguments to arguments for each individual function.
    argument_positions = []
    for func_args in args:
        positions = []
        for arg in func_args:
            positions.append(merged_args.index(arg))
        argument_positions.append(positions)

    logger.debug(f"funcs: {funcs}, args: {args}")
    logger.debug(f"merged args: {merged_args}")
    logger.debug(f"argument_positions: {argument_positions}")

    return merged_args, argument_positions

def _function_arguments_from_argument_positions(argument_positions: T_ArgumentPositions,
                                                *args: Union[float, np.ndarray]
                                                ) -> Iterable[List[Union[float, np.ndarray]]]:
    """ Extract the function arguments from a larger set of arguments given the argument positions.

    It determines the arguments for calling functions stored in ``SimultaneousFit`` or ``AddPDF`` classes. Use it
    something like:

    >>> pdf = AddPDF(f_1, f_2)
    >>> for func, function_args in zip(pdf.functions,
    ...                                _function_arguments_from_argument_positions(
    ...                                    pdf.argument_positions, *full_set_of_args
    ...                                )):
    ...     # Call the function
    ...     func(*function_args)

    Args:
        argument_positions: Map from merged arguments to arguments for each function.
        args: Arguments for the functions. Must include the x argument!
    Returns:
        List of arguments for the function that is described by the maps in argument_positions.
    """
    for arg_positions in argument_positions:
        #logger.debug(f"arg_positions: {arg_positions}, full args: {args}")
        # Determine the arguments for the given function using the arg positions map.
        function_args = []
        for v in arg_positions:
            # We don't skip over the x argument because it's expected to be supplied in the args.
            function_args.append(args[v])
        #logger.debug(f"arg_positions: {arg_positions}, function_args: {function_args}")
        yield function_args

def call_list_of_callables_with_operation(operation: Callable[[float, float], float],
                                          functions: Iterable[Callable[..., float]],
                                          argument_positions: T_ArgumentPositions,
                                          *args: Union[float, np.ndarray]) -> float:
    """ Call and add a list of callables with the given args.

    Args:
        functions: Functions to be evaluated.
        argument_positions: Map from merged arguments to arguments for each function.
        args: Arguments for the functions. Must include the x argument!
    Returns:
        Sum of the values of the functions.
    """
    values = []
    # We write it out explicitly instead of using a generator because it's long enough that it would be difficult
    # to understand.
    for func, function_args in zip(functions, _function_arguments_from_argument_positions(argument_positions, *args)):
        values.append(func(*function_args))
    return functools.reduce(operation, values)

def chi_squared_probability(chi_2: float, ndf: float) -> float:
    """ Calculate the probability that the

    This is just a thin wrapped around ``scipy.stats``, but it's convenient.

    Args:
        chi_2: Chi squared value.
        ndf: Number of degrees of freedom.
    Returns:
        Probability that the fit is consistent with the data.
    """
    import scipy.stats
    return cast(float, 1 - scipy.stats.chi2.cdf(chi_2, ndf))



#!/usr/bin/env python3

""" Tests for functions related to fitting.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import logging
from typing import Any, List

import numpy as np
import pytest

from pachyderm.fit import function

logger = logging.getLogger(__name__)

def test_gaussian(logging_mixin: Any) -> None:
    """ Test the gaussian function. """
    # Setup
    x = np.arange(-10, 10, 0.1)
    mean = 0.5
    sigma = 4.2

    # Calculate independently outside of the tests for mean = 0.5, sigma = 4.2.
    expected = np.array([
        0.0041734 , 0.00442811, 0.00469569, 0.00497663, 0.00527138,  # noqa: E203
        0.00558043, 0.00590424, 0.00624331, 0.0065981 , 0.00696911,  # noqa: E203
        0.0073568 , 0.00776167, 0.00818417, 0.00862478, 0.00908396,  # noqa: E203
        0.00956216, 0.01005984, 0.01057741, 0.01111532, 0.01167396,  # noqa: E203
        0.01225372, 0.01285499, 0.01347812, 0.01412345, 0.01479129,  # noqa: E203
        0.01548193, 0.01619563, 0.01693263, 0.01769313, 0.01847732,  # noqa: E203
        0.01928533, 0.02011726, 0.02097318, 0.02185314, 0.0227571 ,  # noqa: E203
        0.02368503, 0.02463683, 0.02561235, 0.02661141, 0.02763376,  # noqa: E203
        0.02867914, 0.02974718, 0.03083752, 0.03194971, 0.03308325,  # noqa: E203
        0.03423759, 0.03541212, 0.03660619, 0.03781908, 0.03905002,  # noqa: E203
        0.04029816, 0.04156264, 0.04284249, 0.04413673, 0.0454443 ,  # noqa: E203
        0.04676408, 0.04809493, 0.04943561, 0.05078487, 0.05214139,  # noqa: E203
        0.0535038 , 0.05487069, 0.05624062, 0.05761208, 0.05898353,  # noqa: E203
        0.06035341, 0.06172011, 0.06308198, 0.06443737, 0.06578457,  # noqa: E203
        0.06712188, 0.06844755, 0.06975986, 0.07105703, 0.0723373 ,  # noqa: E203
        0.07359891, 0.07484008, 0.07605905, 0.07725407, 0.07842339,  # noqa: E203
        0.0795653 , 0.08067808, 0.08176006, 0.0828096 , 0.08382508,  # noqa: E203
        0.08480492, 0.08574759, 0.0866516 , 0.08751552, 0.08833796,  # noqa: E203
        0.08911759, 0.08985315, 0.09054344, 0.09118732, 0.09178374,  # noqa: E203
        0.0923317 , 0.09283029, 0.09327869, 0.09367612, 0.09402194,  # noqa: E203
        0.09431555, 0.09455646, 0.09474425, 0.09487862, 0.09495934,  # noqa: E203
        0.09498626, 0.09495934, 0.09487862, 0.09474425, 0.09455646,  # noqa: E203
        0.09431555, 0.09402194, 0.09367612, 0.09327869, 0.09283029,  # noqa: E203
        0.0923317 , 0.09178374, 0.09118732, 0.09054344, 0.08985315,  # noqa: E203
        0.08911759, 0.08833796, 0.08751552, 0.0866516 , 0.08574759,  # noqa: E203
        0.08480492, 0.08382508, 0.0828096 , 0.08176006, 0.08067808,  # noqa: E203
        0.0795653 , 0.07842339, 0.07725407, 0.07605905, 0.07484008,  # noqa: E203
        0.07359891, 0.0723373 , 0.07105703, 0.06975986, 0.06844755,  # noqa: E203
        0.06712188, 0.06578457, 0.06443737, 0.06308198, 0.06172011,  # noqa: E203
        0.06035341, 0.05898353, 0.05761208, 0.05624062, 0.05487069,  # noqa: E203
        0.0535038 , 0.05214139, 0.05078487, 0.04943561, 0.04809493,  # noqa: E203
        0.04676408, 0.0454443 , 0.04413673, 0.04284249, 0.04156264,  # noqa: E203
        0.04029816, 0.03905002, 0.03781908, 0.03660619, 0.03541212,  # noqa: E203
        0.03423759, 0.03308325, 0.03194971, 0.03083752, 0.02974718,  # noqa: E203
        0.02867914, 0.02763376, 0.02661141, 0.02561235, 0.02463683,  # noqa: E203
        0.02368503, 0.0227571 , 0.02185314, 0.02097318, 0.02011726,  # noqa: E203
        0.01928533, 0.01847732, 0.01769313, 0.01693263, 0.01619563,  # noqa: E203
        0.01548193, 0.01479129, 0.01412345, 0.01347812, 0.01285499,  # noqa: E203
        0.01225372, 0.01167396, 0.01111532, 0.01057741, 0.01005984,  # noqa: E203
        0.00956216, 0.00908396, 0.00862478, 0.00818417, 0.00776167   # noqa: E203
    ])

    # Calculate using our defined gaussian
    results = function.gaussian(x = x, mean = mean, sigma = sigma)

    # Check the result.
    # We need a bit of extra tolerance because they are calculated at different times to float precision.
    np.testing.assert_allclose(results, expected, atol = 1e-6)

@pytest.mark.parametrize("skip_prefixes, expected_co_varnames", [  # type:ignore
    (["x"], ["x", "f_a", "f_b", "g_c", "g_d"]),
    (["x", "a"], ["x", "a", "f_b", "g_c", "g_d"]),
], ids = ["No skipped prefixes", "Skipped x prefix"])
def test_AddPDF(logging_mixin: Any, simple_test_functions: Any, skip_prefixes: List[str],
                expected_co_varnames: List[str]) -> None:
    """ Test for adding multiple functions with ``AddPDF``.

    Note:
        This is effectively an integration test for most of the ``AddPDF`` features.
    """
    # Setup
    func_1, func_2 = simple_test_functions

    # Create added functions, prepending prefixes, and skipping a subset of the prefixes.
    # This is effectively an integration test.
    added = function.AddPDF(func_1, func_2, prefixes = ["f", "g"], skip_prefixes = skip_prefixes)

    # Check properties
    assert added.func_code.co_varnames == expected_co_varnames
    np.testing.assert_allclose(added(np.array([1, 2]), *[1., 3, 3, 4]), [5 + 8, 6 + 9])

@pytest.mark.parametrize("skip_prefixes, expected_co_varnames", [  # type:ignore
    (["x"], ["x", "f_a", "f_b", "g_c", "g_d"]),
    (["x", "a"], ["x", "a", "f_b", "g_c", "g_d"]),
], ids = ["No skipped prefixes", "Skipped x prefix"])
def test_SubtractPDF(logging_mixin: Any, simple_test_functions: Any, skip_prefixes: List[str],
                     expected_co_varnames: List[str]) -> None:
    """ Test for subtracting multiple functions with ``SubtractPDF``.

    Note:
        This is effectively an integration test for most of the ``SubtractPDF`` features.
    """
    # Setup
    func_1, func_2 = simple_test_functions

    # Create added functions, prepending prefixes, and skipping a subset of the prefixes.
    # This is effectively an integration test.
    subtracted = function.SubtractPDF(func_1, func_2, prefixes = ["f", "g"], skip_prefixes = skip_prefixes)

    # Check properties
    assert subtracted.func_code.co_varnames == expected_co_varnames
    np.testing.assert_allclose(subtracted(np.array([1, 2]), *[1., 3, 3, 4]), [5. - 8, 6. - 9])

@pytest.mark.parametrize("skip_prefixes, expected_co_varnames", [  # type:ignore
    (["x"], ["x", "f_a", "f_b", "g_c", "g_d"]),
    (["x", "a"], ["x", "a", "f_b", "g_c", "g_d"]),
], ids = ["No skipped prefixes", "Skipped x prefix"])
def test_MultiplyPDF(logging_mixin: Any, simple_test_functions: Any, skip_prefixes: List[str],
                     expected_co_varnames: List[str]) -> None:
    """ Test for multiplying multiple functions with ``MultiplyPDF``.

    Note:
        This is effectively an integration test for most of the ``MultiplyPDF`` features.
    """
    # Setup
    func_1, func_2 = simple_test_functions

    # Create added functions, prepending prefixes, and skipping a subset of the prefixes.
    # This is effectively an integration test.
    multiplied = function.MultiplyPDF(func_1, func_2, prefixes = ["f", "g"], skip_prefixes = skip_prefixes)

    # Check properties
    assert multiplied.func_code.co_varnames == expected_co_varnames
    np.testing.assert_allclose(multiplied(np.array([1, 2]), *[1., 3, 3, 4]), [5. * 8, 6. * 9])

@pytest.mark.parametrize("skip_prefixes, expected_co_varnames", [  # type:ignore
    (["x"], ["x", "f_a", "f_b", "g_c", "g_d"]),
    (["x", "a"], ["x", "a", "f_b", "g_c", "g_d"]),
], ids = ["No skipped prefixes", "Skipped x prefix"])
def test_DividePDF(logging_mixin: Any, simple_test_functions: Any, skip_prefixes: List[str],
                   expected_co_varnames: List[str]) -> None:
    """ Test for dividing multiple functions with ``DividePDF``.

    Note:
        This is effectively an integration test for most of the ``DividePDF`` features.
    """
    # Setup
    func_1, func_2 = simple_test_functions

    # Create added functions, prepending prefixes, and skipping a subset of the prefixes.
    # This is effectively an integration test.
    divided = function.DividePDF(func_1, func_2, prefixes = ["f", "g"], skip_prefixes = skip_prefixes)

    # Check properties
    assert divided.func_code.co_varnames == expected_co_varnames
    np.testing.assert_allclose(divided(np.array([1, 2]), *[1., 3, 3, 4]), [5. / 8, 6. / 9])



#!/usr/bin/env python

""" Typing helpers for package.

.. codeauthor:: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

from typing import Any, Union

try:
    import ROOT
    # Tell ROOT to ignore command line options so args are passed to python
    # NOTE: Must be immediately after import ROOT and it must be called the first time ROOT is imported!
    #       We do this here (even though it is unrelated to typing helpers) because it is the most common
    #       first import which requires ROOT. So by putting it here, it should (hopefully) cover all executables.
    #       However, this means that it needs to be imported before some pachyderm modules.
    ROOT.PyConfig.IgnoreCommandLineOptions = True

    Hist = Union[ROOT.TH1, ROOT.THnBase]
    Axis = ROOT.TAxis
    Canvas = ROOT.TCanvas
    TFile = ROOT.TFile
except ImportError:
    # It doesn't like the possibility of redefining this, so we need to tell ``mypy`` to ignore it.
    Hist = Any  # type: ignore
    Axis = Any
    Canvas = Any
    TFile = Any



#!/usr/bin/env python3

""" Functions for use with fitting.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import abc
import logging
import operator
from typing import Callable, Optional, Sequence, Union

import numpy as np

from pachyderm import generic_class
from pachyderm.fit import base as fit_base

logger = logging.getLogger(__name__)

class CombinePDF(generic_class.EqualityMixin, abc.ABC):
    """ Combine functions (PDFs) together.

    Args:
        functions: Functions to be added.
        prefixes: Prefix for arguments of each function. Default: None. If specified, there must
            be one prefix for each function.
        skip_prefixes: Prefixes to skip when assigning prefixes. As noted in probfit, this can be
            useful to mix prefixed and non-prefixed arguments. Default: None.

    Attributes:
        functions: List of functions that are combined in the PDF.
        func_code: Function arguments derived from the fit functions. They need to be separately
            specified to allow iminuit to determine the proper arguments.
        argument_positions: Map of merged arguments to the arguments for each individual function.
    """
    # Don't specify the function arguments to work aorund a mypy bug.
    # For an unclear reason, it won't properly detect the number of arguments.
    _operation: Callable[..., float]
    _call_function: Callable[..., float]

    def __init__(self, *functions: Callable[..., float], prefixes: Optional[Sequence[str]] = None,
                 skip_prefixes: Optional[Sequence[str]] = None) -> None:
        # Store the functions
        self.functions = list(functions)

        # Determine the arguments for the functions.
        merged_args, argument_positions = fit_base.merge_func_codes(
            self.functions, prefixes = prefixes, skip_prefixes = skip_prefixes
        )
        logger.debug(f"merged_args: {merged_args}")
        self.func_code = fit_base.FuncCode(merged_args)
        self.argument_positions = argument_positions

    def __call__(self, x: np.ndarray, *merged_args: float) -> float:
        """ Call the added PDF.

        Args:
            x: Value(s) where the functions should be evaluated.
            merged_args: Merged arguments for the functions. Must contain all of the arguments
                need to call the functions.
        Returns:
            Value(s) of the functions when evaluated with the given input values.
        """
        # We add in the x values into the function arguments here so we don't have to play tricks later
        # to get the function argumment indices correct.
        return fit_base.call_list_of_callables_with_operation(
            self._operation,
            self.functions, self.argument_positions, *[x, *merged_args]
        )

class AddPDF(CombinePDF):
    """ Add functions (PDFs) together.

    Args:
        functions: Functions to be added.
        prefixes: Prefix for arguments of each function. Default: None. If specified, there must
            be one prefix for each function.
        skip_prefixes: Prefixes to skip when assigning prefixes. As noted in probfit, this can be
            useful to mix prefixed and non-prefixed arguments. Default: None.

    Attributes:
        functions: List of functions that are combined in the PDF.
        func_code: Function arguments derived from the fit functions. They need to be separately
            specified to allow iminuit to determine the proper arguments.
        argument_positions: Map of merged arguments to the arguments for each individual function.
    """
    _operation = operator.add

class SubtractPDF(CombinePDF):
    """ Subtract functions (PDFs) together.

    Args:
        functions: Functions to be added.
        prefixes: Prefix for arguments of each function. Default: None. If specified, there must
            be one prefix for each function.
        skip_prefixes: Prefixes to skip when assigning prefixes. As noted in probfit, this can be
            useful to mix prefixed and non-prefixed arguments. Default: None.

    Attributes:
        functions: List of functions that are combined in the PDF.
        func_code: Function arguments derived from the fit functions. They need to be separately
            specified to allow iminuit to determine the proper arguments.
        argument_positions: Map of merged arguments to the arguments for each individual function.
    """
    _operation = operator.sub

class MultiplyPDF(CombinePDF):
    """ Multiply functions (PDFs) together.

    Args:
        functions: Functions to be added.
        prefixes: Prefix for arguments of each function. Default: None. If specified, there must
            be one prefix for each function.
        skip_prefixes: Prefixes to skip when assigning prefixes. As noted in probfit, this can be
            useful to mix prefixed and non-prefixed arguments. Default: None.

    Attributes:
        functions: List of functions that are combined in the PDF.
        func_code: Function arguments derived from the fit functions. They need to be separately
            specified to allow iminuit to determine the proper arguments.
        argument_positions: Map of merged arguments to the arguments for each individual function.
    """
    _operation = operator.mul

class DividePDF(CombinePDF):
    """ Divide functions (PDFs) together.

    Args:
        functions: Functions to be added.
        prefixes: Prefix for arguments of each function. Default: None. If specified, there must
            be one prefix for each function.
        skip_prefixes: Prefixes to skip when assigning prefixes. As noted in probfit, this can be
            useful to mix prefixed and non-prefixed arguments. Default: None.

    Attributes:
        functions: List of functions that are combined in the PDF.
        func_code: Function arguments derived from the fit functions. They need to be separately
            specified to allow iminuit to determine the proper arguments.
        argument_positions: Map of merged arguments to the arguments for each individual function.
    """
    _operation = operator.truediv

def gaussian(x: Union[np.ndarray, float], mean: float, sigma: float) -> Union[np.ndarray, float]:
    r""" Normalized gaussian.

    .. math::

        f = 1 / \sqrt{2 * \pi * \sigma^{2}} * \exp{-\frac{(x - \mu)^{2}}{(2 * \sigma^{2}}}

    Args:
        x: Value(s) where the gaussian should be evaluated.
        mean: Mean of the gaussian distribution.
        sigma: Width of the gaussian distribution.
    Returns:
        Calculated gaussian value(s).
    """
    return 1.0 / np.sqrt(2 * np.pi * np.square(sigma)) * np.exp(-1.0 / 2.0 * np.square((x - mean) / sigma))

def extended_gaussian(x: Union[np.ndarray, float], mean: float, sigma: float,
                      amplitude: float) -> Union[np.ndarray, float]:
    r""" Extended gaussian.

    .. math::

        f = A / \sqrt{2 * \pi * \sigma^{2}} * \exp{-\frac{(x - \mu)^{2}}{(2 * \sigma^{2}}}

    Args:
        x: Value(s) where the gaussian should be evaluated.
        mean: Mean of the gaussian distribution.
        sigma: Width of the gaussian distribution.
        amplitude: Amplitude of the gaussian.
    Returns:
        Calculated gaussian value(s).
    """
    return amplitude / np.sqrt(2 * np.pi * np.square(sigma)) * np.exp(-1.0 / 2.0 * np.square((x - mean) / sigma))



#!/usr/bin/env python3

""" Tests for integration of functionality in the fit modules.

.. code-author: Raymond Ehlers <raymond.ehlers@cern.ch>, Yale University
"""

import logging
from io import StringIO
from typing import Any, Tuple, Union

import numpy as np
import pytest  # noqa: F401

import pachyderm.fit
from pachyderm import histogram, yaml

logger = logging.getLogger(__name__)

def dump_to_string_and_retrieve(input_object: Any, y: yaml.ruamel.yaml.YAML = None) -> Any:
    """ Dump the given input object via YAML and then retrieve it for comparison.

    Args:
        input_object: Object to be dumped and retrieved.
        y: YAML object to use for the dumping. If not specified, one will be created.
    Returns:
        The dumped and then retrieved object.
    """
    # Create a YAML object if necessary
    if y is None:
        y = yaml.yaml()

    # Dump to a string
    s = StringIO()
    y.dump([input_object], s)
    s.seek(0)
    # And then load from the string. Note the implicit unpacking
    output_object, = y.load(s)

    return output_object

def pedestal(x: Union[np.ndarray, float], pedestal: float) -> Union[np.ndarray]:
    return pedestal

class PedestalFit(pachyderm.fit.Fit):
    """ Minimal pedestal fit object to test reading and writing to YAML. """
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.fit_function = pedestal

    def _post_init_validation(self) -> None:
        """ Validate that the fit object was setup properly. """
        ...

    def _setup(self, h: histogram.Histogram1D) -> Tuple[histogram.Histogram1D, pachyderm.fit.T_FitArguments]:
        """ Setup the histogram and arguments for the fit. """
        return h, {"pedestal": 0}

def test_round_trip_of_fit_object_to_YAML(logging_mixin: Any) -> None:
    """ Test a YAML round trip for a fit object. """
    # Setup
    input_fit_object = PedestalFit(use_log_likelihood = False)
    # YAML object
    y = yaml.yaml(classes_to_register = [PedestalFit])

    # Dump and load (ie round trip)
    output_fit_object = dump_to_string_and_retrieve(input_fit_object, y = y)

    # Check the result
    assert output_fit_object == input_fit_object
